# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test to make sure the `TensorSpace` spaces work with larger sizes."""

from __future__ import division

import numpy as np
import pytest

import odl
from odl.core.util.dtype_utils import _universal_dtype_identifier
from odl.core.util.testutils import all_almost_equal, dtype_tol, noise_elements

# --- pytest fixtures --- #


# Global pytest mark
pytestmark = pytest.mark.suite('largescale')


spc_params = ['rn', '1d', '3d']
spc_ids = [' type={} '.format(p) for p in spc_params]


@pytest.fixture(scope="module", ids=spc_ids, params=spc_params)
def tspace(odl_impl_device_pairs, request):
    spc = request.param
    impl, device = odl_impl_device_pairs

    if spc == 'rn':
        return odl.rn(10 ** 5, impl=impl, device=device)
    elif spc == '1d':
        return odl.uniform_discr(0, 1, 10 ** 5, impl=impl, device=device)
    elif spc == '3d':
        return odl.uniform_discr([0, 0, 0], [1, 1, 1],
                                 [100, 100, 100], impl=impl, device=device)


def test_element(tspace):
    x = tspace.element()
    assert x in tspace

    # From array-like
    y = tspace.element(np.zeros(tspace.shape, dtype=_universal_dtype_identifier(tspace.dtype)).tolist())
    assert y in tspace

    # Rewrap
    y2 = tspace.element(y)
    assert y2 is y

    w = tspace.element(np.zeros(tspace.shape, dtype=_universal_dtype_identifier(tspace.dtype)))
    assert w in tspace


def test_zero(tspace):
    assert all_almost_equal(tspace.zero(), 0)


def test_one(tspace):
    assert all_almost_equal(tspace.one(), 1)

def test_ndarray_init(tspace):
    x0 = tspace.array_namespace.arange(tspace.size, device=tspace.device).reshape(tspace.shape)
    x = tspace.element(x0)

    assert all_almost_equal(x0, x)


def test_getitem(tspace):
    indices = np.random.randint(0, tspace.size - 1, 5)
    indices = np.unravel_index(indices, tspace.shape)

    x0 = np.arange(tspace.size).reshape(tspace.shape)
    x = tspace.element(x0)

    for index in zip(*indices):
        assert x[index] == np.ravel_multi_index(index, tspace.shape)


def test_setitem(tspace):
    indices = np.random.randint(0, tspace.size - 1, 5)
    indices = np.unravel_index(indices, tspace.shape)

    x = tspace.zero()

    for index in zip(*indices):
        flat_index = np.ravel_multi_index(index, tspace.shape)
        x[index] = -flat_index
        assert x[index] == -flat_index


def test_inner(tspace):
    weighting_const = tspace.weighting.const

    [xarr, yarr], [x, y] = noise_elements(tspace, 2)

    correct_inner = tspace.array_namespace.vdot(yarr.ravel(), xarr.ravel()) * weighting_const

    assert (
        tspace.inner(x, y)
        == pytest.approx(float(correct_inner), rel=dtype_tol(tspace.dtype))
    )
    assert (
        x.inner(y)
        == pytest.approx(float(correct_inner), rel=dtype_tol(tspace.dtype))
    )


def test_norm(tspace):
    weighting_const = tspace.weighting.const

    xarr, x = noise_elements(tspace)

    correct_norm = tspace.array_namespace.linalg.norm(xarr) * np.sqrt(weighting_const)

    assert (
        tspace.norm(x)
        == pytest.approx(float(correct_norm), rel=dtype_tol(tspace.dtype))
    )
    assert (
        x.norm()
        == pytest.approx(float(correct_norm), rel=dtype_tol(tspace.dtype))
    )


def test_dist(tspace):
    weighting_const = tspace.weighting.const

    [xarr, yarr], [x, y] = noise_elements(tspace, 2)

    correct_dist = tspace.array_namespace.linalg.norm(xarr - yarr) * np.sqrt(weighting_const)

    assert (
        tspace.dist(x, y)
        == pytest.approx(float(correct_dist), rel=dtype_tol(tspace.dtype))
    )
    assert (
        x.dist(y)
        == pytest.approx(float(correct_dist), rel=dtype_tol(tspace.dtype))
    )


def _test_lincomb(space, a, b, discontig):
    """Validate lincomb against direct result using arrays."""
    # Set slice for discontiguous arrays and get result space of slicing
    if discontig:
        slc = tuple([slice(None)] * (space.ndim - 1) + [slice(None, None, 2)])
        res_space = space.element()[slc].space
    else:
        res_space = space

    # Unaliased arguments
    [xarr, yarr, zarr], [x, y, z] = noise_elements(space, 3)
    if discontig:
        x, y, z = x[slc], y[slc], z[slc]
        xarr, yarr, zarr = xarr[slc], yarr[slc], zarr[slc]

    zarr[:] = a * xarr + b * yarr
    res_space.lincomb(a, x, b, y, out=z)
    assert all_almost_equal([x, y, z], [xarr, yarr, zarr])

    # First argument aliased with output
    [xarr, yarr, zarr], [x, y, z] = noise_elements(space, 3)
    if discontig:
        x, y, z = x[slc], y[slc], z[slc]
        xarr, yarr, zarr = xarr[slc], yarr[slc], zarr[slc]

    zarr[:] = a * zarr + b * yarr
    res_space.lincomb(a, z, b, y, out=z)
    assert all_almost_equal([x, y, z], [xarr, yarr, zarr])

    # Second argument aliased with output
    [xarr, yarr, zarr], [x, y, z] = noise_elements(space, 3)
    if discontig:
        x, y, z = x[slc], y[slc], z[slc]
        xarr, yarr, zarr = xarr[slc], yarr[slc], zarr[slc]

    zarr[:] = a * xarr + b * zarr
    res_space.lincomb(a, x, b, z, out=z)
    assert all_almost_equal([x, y, z], [xarr, yarr, zarr])

    # Both arguments aliased with each other
    [xarr, yarr, zarr], [x, y, z] = noise_elements(space, 3)
    if discontig:
        x, y, z = x[slc], y[slc], z[slc]
        xarr, yarr, zarr = xarr[slc], yarr[slc], zarr[slc]

    zarr[:] = a * xarr + b * xarr
    res_space.lincomb(a, x, b, x, out=z)
    assert all_almost_equal([x, y, z], [xarr, yarr, zarr])

    # All aliased
    [xarr, yarr, zarr], [x, y, z] = noise_elements(space, 3)
    if discontig:
        x, y, z = x[slc], y[slc], z[slc]
        xarr, yarr, zarr = xarr[slc], yarr[slc], zarr[slc]

    zarr[:] = a * zarr + b * zarr
    res_space.lincomb(a, z, b, z, out=z)
    assert all_almost_equal([x, y, z], [xarr, yarr, zarr])


def test_lincomb(tspace):
    scalar_values = [0, 1, -1, 3.41]
    for a in scalar_values:
        for b in scalar_values:
            _test_lincomb(tspace, a, b, discontig=False)
            _test_lincomb(tspace, a, b, discontig=True)


def _test_member_lincomb(spc, a):
    # Validates vector member lincomb against the result on host

    # Generate vectors
    [x_host, y_host], [x_device, y_device] = noise_elements(spc, 2)

    # Host side calculation
    y_host[:] = a * x_host

    # Device side calculation
    y_device.lincomb(a, x_device)

    # CUDA only uses floats, so require 2 digits
    assert all_almost_equal(y_device, y_host, ndigits=2)


def test_member_lincomb(tspace):
    scalar_values = [0, 1, -1, 3.41, 10.0, 1.0001]
    for a in scalar_values:
        _test_member_lincomb(tspace, a)


def _test_unary_operator(spc, function):
    # Verify that the statement y=function(x) gives equivalent
    # results to Numpy.
    x_arr, x = noise_elements(spc)

    y_arr = function(x_arr)
    y = function(x)

    assert all_almost_equal([x, y],
                            [x_arr, y_arr])


def _test_binary_operator(spc, function):
    # Verify that the statement z=function(x,y) gives equivalent
    # results to Numpy.
    [x_arr, y_arr], [x, y] = noise_elements(spc, 2)

    z_arr = function(x_arr, y_arr)
    z = function(x, y)

    assert all_almost_equal([x, y, z],
                            [x_arr, y_arr, z_arr])


def test_operators(tspace):
    # Test of all operator overloads against the corresponding
    # Numpy implementation

    # Unary operators
    _test_unary_operator(tspace, lambda x: +x)
    _test_unary_operator(tspace, lambda x: -x)

    # Scalar multiplication
    for scalar in [-31.2, -1, 0, 1, 2.13]:
        def imul(x):
            x *= scalar
        _test_unary_operator(tspace, imul)
        _test_unary_operator(tspace, lambda x: x * scalar)

    # Scalar division
    for scalar in [-31.2, -1, 1, 2.13]:
        def idiv(x):
            x /= scalar
        _test_unary_operator(tspace, idiv)
        _test_unary_operator(tspace, lambda x: x / scalar)

    # Incremental operations
    def iadd(x, y):
        x += y

    def isub(x, y):
        x -= y

    def imul(x, y):
        x *= y

    def idiv(x, y):
        x /= y

    _test_binary_operator(tspace, iadd)
    _test_binary_operator(tspace, isub)
    _test_binary_operator(tspace, imul)
    _test_binary_operator(tspace, idiv)

    # Incremental operators with aliased inputs
    def iadd_aliased(x):
        x += x

    def isub_aliased(x):
        x -= x

    def imul_aliased(x):
        x *= x

    def idiv_aliased(x):
        x /= x

    _test_unary_operator(tspace, iadd_aliased)
    _test_unary_operator(tspace, isub_aliased)
    _test_unary_operator(tspace, imul_aliased)
    _test_unary_operator(tspace, idiv_aliased)

    # Binary operators
    _test_binary_operator(tspace, lambda x, y: x + y)
    _test_binary_operator(tspace, lambda x, y: x - y)
    _test_binary_operator(tspace, lambda x, y: x * y)
    _test_binary_operator(tspace, lambda x, y: x / y)

    # Binary with aliased inputs
    _test_unary_operator(tspace, lambda x: x + x)
    _test_unary_operator(tspace, lambda x: x - x)
    _test_unary_operator(tspace, lambda x: x * x)
    _test_unary_operator(tspace, lambda x: x / x)


if __name__ == '__main__':
    odl.core.util.test_file(__file__, ['-S', 'largescale'])



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Utilities for datasets."""

from __future__ import print_function
import os
from os.path import join, expanduser, exists
from future.moves.urllib.request import urlopen
from shutil import copyfileobj, rmtree
from scipy import io
import contextlib


__all__ = ('get_data_dir', 'cleanup_data_dir', 'get_data')


def get_data_dir():
    """Get the data directory."""
    base_odl_dir = os.environ.get('ODL_HOME', expanduser(join('~', '.odl')))
    data_home = join(base_odl_dir, 'datasets')
    if not exists(data_home):
        os.makedirs(data_home)
    return data_home


def cleanup_data_dir():
    """Remove all data, forcing it to be reloaded."""
    rmtree(get_data_dir())


def get_data(filename, subset, url):
    """Get a dataset with from a url with local caching.

    Parameters
    ----------
    filename : str
        Name of the file, for caching.
    subset : str
        To what subset the file belongs (e.g. 'ray_transform'). Each subset
        is saved in a separate subfolder.
    url : str
        url to the dataset online.

    Returns
    -------
    dataset : dict
        Dictionary containing the dataset.
    """
    # check if this data set has been already downloaded
    data_dir = join(get_data_dir(), subset)
    if not exists(data_dir):
        os.makedirs(data_dir)

    filename = join(data_dir, filename)

    # if the file does not exist, download it
    if not exists(filename):
        print('data {}/{} not in local storage, downloading from {}'
              ''.format(subset, filename, url))

        # open the url of the data
        with contextlib.closing(urlopen(url)) as data_url:
            # store downloaded file locally
            with open(filename, 'w+b') as storage_file:
                copyfileobj(data_url, storage_file)

    # load dataset file
    with open(filename, 'rb') as storage_file:
        data_dict = io.loadmat(storage_file)

    return data_dict


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division

import pytest

import odl
from odl.core.util.testutils import (
    all_almost_equal, noise_element, simple_fixture, skip_if_no_pywavelets)


# --- pytest fixtures --- #


wavelet = simple_fixture('wavelet', ['db1', 'sym2'])
pad_mode = simple_fixture('pad_mode', ['constant', 'pywt_periodic'])
ndim = simple_fixture('ndim', [1, 2, 3])
nlevels = simple_fixture('nlevels', [2, None])
axes = simple_fixture('axes', [-1, None])
wave_impl = simple_fixture(
    'wave_impl',
    [pytest.param('pywt', marks=skip_if_no_pywavelets)]
)


@pytest.fixture(scope='module')
def shape_setup(ndim, nlevels, wavelet, pad_mode):
    if ndim == 1:
        image_shape = (16,)
        if wavelet == 'db1':
            coeff_shapes = [(4,), (4,), (8,)]
        elif wavelet == 'sym2':
            if pad_mode == 'constant':
                coeff_shapes = [(6,), (6,), (9,)]
            elif pad_mode == 'pywt_periodic':
                coeff_shapes = [(4,), (4,), (8,)]
            else:
                raise RuntimeError
        else:
            raise RuntimeError

    elif ndim == 2:
        image_shape = (16, 17)
        if wavelet == 'db1':
            coeff_shapes = [(4, 5), (4, 5), (8, 9)]
        elif wavelet == 'sym2':
            if pad_mode == 'constant':
                coeff_shapes = [(6, 6), (6, 6), (9, 10)]
            elif pad_mode == 'pywt_periodic':
                coeff_shapes = [(4, 5), (4, 5), (8, 9)]
            else:
                raise RuntimeError
        else:
            raise RuntimeError

    elif ndim == 3:
        image_shape = (16, 17, 18)
        if wavelet == 'db1':
            coeff_shapes = [(4, 5, 5), (4, 5, 5), (8, 9, 9)]
        elif wavelet == 'sym2':
            if pad_mode == 'constant':
                coeff_shapes = [(6, 6, 6), (6, 6, 6), (9, 10, 10)]
            elif pad_mode == 'pywt_periodic':
                coeff_shapes = [(4, 5, 5), (4, 5, 5), (8, 9, 9)]
            else:
                raise RuntimeError
    else:
        raise RuntimeError

    return wavelet, pad_mode, nlevels, image_shape, coeff_shapes


def test_wavelet_transform(wave_impl, shape_setup, odl_floating_dtype, axes):
    # Verify that the operator works as expected
    dtype = odl_floating_dtype
    wavelet, pad_mode, nlevels, shape, _ = shape_setup
    ndim = len(shape)

    space = odl.uniform_discr([-1] * ndim, [1] * ndim, shape, dtype=dtype)
    image = noise_element(space)

    # TODO: check more error scenarios
    if wave_impl == 'pywt' and pad_mode == 'constant':
        with pytest.raises(ValueError):
            wave_trafo = odl.trafos.WaveletTransform(
                space, wavelet, nlevels, pad_mode, pad_const=1, impl=wave_impl,
                axes=axes)

    wave_trafo = odl.trafos.WaveletTransform(
        space, wavelet, nlevels, pad_mode, impl=wave_impl, axes=axes)

    assert wave_trafo.domain.dtype == dtype
    assert wave_trafo.range.dtype == dtype

    wave_trafo_inv = wave_trafo.inverse
    assert wave_trafo_inv.domain.dtype == dtype
    assert wave_trafo_inv.range.dtype == dtype
    assert wave_trafo_inv.nlevels == wave_trafo.nlevels
    assert wave_trafo_inv.wavelet == wave_trafo.wavelet
    assert wave_trafo_inv.pad_mode == wave_trafo.pad_mode
    assert wave_trafo_inv.pad_const == wave_trafo.pad_const
    assert wave_trafo_inv.pywt_pad_mode == wave_trafo.pywt_pad_mode

    coeffs = wave_trafo(image)
    reco_image = wave_trafo.inverse(coeffs)
    assert all_almost_equal(image.real, reco_image.real)
    assert all_almost_equal(image, reco_image)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



from dataclasses import dataclass
from typing import Callable

_registered_sparse_formats = {}

@dataclass
class SparseMatrixFormat:
    sparse_format : str
    impl : str
    constructor : Callable
    is_of_this_sparse_format : Callable[[object], bool]
    to_dense : Callable
    matmul_spmatrix_with_vector : Callable
    def __post_init__(self):
        if self.impl not in _registered_sparse_formats:
            _registered_sparse_formats[self.impl] = {}
        if self.sparse_format in _registered_sparse_formats[self.impl]:
            raise KeyError(f"A {self.sparse_format} sparse format for backend {self.impl} is already registered. Every sparse format needs to have a unique identifier combination.")
        _registered_sparse_formats[self.impl][self.sparse_format] = self



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test objects for tomography problems."""

from __future__ import absolute_import

from . import phantom_utils
from .emission import *
from .geometric import *
from .misc_phantoms import *
from .noise import *
from .transmission import *

__all__ = ()
__all__ += emission.__all__
__all__ += geometric.__all__
__all__ += misc_phantoms.__all__
__all__ += noise.__all__
__all__ += transmission.__all__



import pytest

import odl

from odl.core.array_API_support.comparisons  import odl_all_equal

from odl.core.util.pytest_config import IMPL_DEVICE_PAIRS
from odl.core.util.testutils import (
    noise_elements, simple_fixture)

DEFAULT_SHAPE = (4,4)

keepdims_function = simple_fixture(
    'keepdims', 
    ['max',
    'mean',
    'min',
    'prod',
    'std',
    'sum',
    'var' ]
    )

cumulative_function = simple_fixture(
    'cumulative', 
    ['cumulative_prod',
    'cumulative_sum']
    )

keepdims = simple_fixture(
    'keepdims', 
    [True, False]
    )

axis = simple_fixture(
    'axis', 
    [0, 1]
    )


@pytest.fixture(scope='module', params=IMPL_DEVICE_PAIRS)
def float_tspace(request, odl_real_floating_dtype):
    impl, device = request.param
    return odl.tensor_space(
        shape=DEFAULT_SHAPE, 
        dtype=odl_real_floating_dtype, 
        impl=impl, 
        device=device
    )

def test_keepdims_function(float_tspace, keepdims_function, keepdims):
    ns = float_tspace.array_namespace
    arr_fn = getattr(ns, keepdims_function)
    odl_fn = getattr(odl, keepdims_function)

    x_arr, x = noise_elements(float_tspace, 1)
    y = odl_fn(x, keepdims=keepdims)
    y_arr = arr_fn(x_arr, keepdims=keepdims)
    assert odl_all_equal(y, y_arr)

def test_cumulative_function(float_tspace, cumulative_function, axis):
    ns = float_tspace.array_namespace
    arr_fn = getattr(ns, cumulative_function)
    odl_fn = getattr(odl, cumulative_function)

    x_arr, x = noise_elements(float_tspace, 1)
    y = odl_fn(x, axis=axis)
    y_arr = arr_fn(x_arr, axis=axis)
    assert odl_all_equal(y, y_arr)



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Numpy functionality that is not uniform across the supported versions.
"""

import numpy as np

# This is supposed to be used as the `copy=AVOID_UNNECESSARY_COPY` argument to
# the `np.array` constructor when a copy is not desired but may not be avoidable
# (for example because a conversion to a different `dtype` is performed).
# ODL uses this in many places, since it gives usually good performance but
# also still offers flexibility.
# In NumPy-1, the `copy` argument could only be `True` or `False`, the latter
# being merely a request which NumPy-1 ignored if necessary.
# By contrast, NumPy-2 makes `False` binding: if a copy cannot be avoided,
# an error is raised. For the old weak request, NumPy-2 offers `copy=None`
# as the value, which is thus what ODL shall use forward-facing.
# NumPy-1 does however not understand this, which is why the following definition
# is needed for compatibility with both.
AVOID_UNNECESSARY_COPY = None if np.__version__>='2' else False

__all__ = ("AVOID_UNNECESSARY_COPY",)

if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Example functionals used in optimization."""

from __future__ import print_function, division, absolute_import
import numpy as np

from odl.functionals.functional import Functional
from odl.core.operator import Operator, MatrixOperator
from odl.core.space.base_tensors import TensorSpace


__all__ = ('RosenbrockFunctional',)


class RosenbrockFunctional(Functional):

    r"""The well-known Rosenbrock function on ``R^n``.

    The `Rosenbrock function`_ is often used as a test problem in
    smooth optimization.

    Notes
    -----
    The functional is defined for :math:`x \\in \\mathbb{R}^n`,
    :math:`n \\geq 2`, as

    .. math::
        \sum_{i=1}^{n - 1} c (x_{i+1} - x_i^2)^2 + (1 - x_i)^2,

    where :math:`c` is a constant, usually set to 100, which determines how
    "ill-behaved" the function should be.
    The global minimum lies at :math:`x = (1, \\dots, 1)`, independent
    of :math:`c`.

    There are two definitions of the n-dimensional Rosenbrock function found in
    the literature. One is the product of 2-dimensional Rosenbrock functions,
    which is not the one used here. This one extends the pattern of the 2d
    Rosenbrock function so all dimensions depend on each other in sequence.

    References
    ----------
    .. _Rosenbrock function: https://en.wikipedia.org/wiki/Rosenbrock_function
    """

    def __init__(self, space, scale=100.0):
        """Initialize a new instance.

        Parameters
        ----------
        space : `TensorSpace`
            Domain of the functional.
        scale : positive float, optional
            The scale ``c`` in the functional determining how
            "ill-behaved" the functional should be. Larger value means
            worse behavior.

        Examples
        --------
        Initialize and call the functional:

        >>> r2 = odl.rn(2)
        >>> functional = RosenbrockFunctional(r2)
        >>> functional([1, 1])  # optimum is 0 at [1, 1]
        0.0
        >>> functional([0, 1])
        101.0

        The functional can also be used in higher dimensions:

        >>> r5 = odl.rn(5)
        >>> functional = RosenbrockFunctional(r5)
        >>> functional([1, 1, 1, 1, 1])
        0.0

        We can change how much the function is ill-behaved via ``scale``:

        >>> r2 = odl.rn(2)
        >>> functional = RosenbrockFunctional(r2, scale=2)
        >>> functional([1, 1])  # optimum is still 0 at [1, 1]
        0.0
        >>> functional([0, 1])  # much lower variation
        3.0
        """
        self.scale = float(scale)
        if not isinstance(space, TensorSpace):
            raise ValueError('`space` must be a `TensorSpace` instance, '
                             'got {!r}'.format(space))
        if space.ndim > 1:
            raise ValueError('`space` cannot have more than 1 dimension')
        if space.size < 2:
            raise ValueError('`space.size` must be >= 2, got {}'
                             ''.format(space.size))
        super(RosenbrockFunctional, self).__init__(
            space, linear=False, grad_lipschitz=np.inf)

    def _call(self, x):
        """Return ``self(x)``."""
        result = 0
        for i in range(0, self.domain.size - 1):
            result += (self.scale * (x[i + 1] - x[i] ** 2) ** 2 +
                       (x[i] - 1) ** 2)

        return result

    @property
    def gradient(self):
        """Gradient operator of the Rosenbrock functional."""
        functional = self
        c = self.scale

        class RosenbrockGradient(Operator):

            """The gradient operator of the Rosenbrock functional."""

            def __init__(self):
                """Initialize a new instance."""
                super(RosenbrockGradient, self).__init__(
                    functional.domain, functional.domain, linear=False)

            def _call(self, x, out):
                """Apply the gradient operator to the given point."""
                for i in range(1, self.domain.size - 1):
                    out[i] = (2 * c * (x[i] - x[i - 1]**2) -
                              4 * c * (x[i + 1] - x[i]**2) * x[i] -
                              2 * (1 - x[i]))
                out[0] = (-4 * c * (x[1] - x[0] ** 2) * x[0] +
                          2 * (x[0] - 1))
                out[-1] = 2 * c * (x[-1] - x[-2] ** 2)

            def derivative(self, x):
                """The derivative of the gradient.

                This is also known as the Hessian.
                """
                # TODO: Implement optimized version of this that does not need
                # a matrix.
                shape = (functional.domain.size, functional.domain.size)
                matrix = np.zeros(shape)

                # Straightforward computation
                for i in range(0, self.domain.size - 1):
                    matrix[i, i] = (2 * c + 2 + 12 * c * x[i] ** 2 -
                                    4 * c * x[i + 1])
                    matrix[i + 1, i] = -4 * c * x[i]
                    matrix[i, i + 1] = -4 * c * x[i]
                matrix[-1, -1] = 2 * c
                matrix[0, 0] = 2 + 12 * c * x[0] ** 2 - 4 * c * x[1]
                return MatrixOperator(matrix, self.domain, self.range)

        return RosenbrockGradient()


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2018 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Partitons of interval products based on rectilinear grids.

A partition of a set is a finite collection of nonempty, pairwise
disjoint subsets whose union is the original set. The partitions
considered here are based on hypercubes, i.e. the tensor products
of partitions of intervals.
"""

from __future__ import print_function, division, absolute_import
from builtins import object
import numpy as np

from odl.core.util.npy_compat import AVOID_UNNECESSARY_COPY

from odl.core.discr.grid import RectGrid, uniform_grid_fromintv
from odl.core.set import IntervalProd
from odl.core.util import (
    normalized_index_expression, normalized_nodes_on_bdry,
    normalized_scalar_param_list, safe_int_conv,
    signature_string, indent, array_str, npy_printoptions)


__all__ = ('RectPartition', 'uniform_partition_fromintv',
           'uniform_partition_fromgrid', 'uniform_partition',
           'nonuniform_partition')


class RectPartition(object):

    """Rectangular partition by hypercubes based on `RectGrid`.

    In 1d, a partition of an interval is implicitly defined by a
    collection of points x[0], ..., x[N-1] (a grid) which are chosen to
    lie in the center of the subintervals. The i-th subinterval is thus
    given by

        ``I[i] = [(x[i-1]+x[i])/2, (x[i]+x[i+1])/2]``

    """

    def __init__(self, intv_prod, grid):
        """Initialize a new instance.

        Parameters
        ----------
        intv_prod : `IntervalProd`
            Set to be partitioned
        grid : `RectGrid`
            Spatial points supporting the partition. They must be
            contained in ``intv_prod``.
        """
        super(RectPartition, self).__init__()

        if not isinstance(intv_prod, IntervalProd):
            raise TypeError('{!r} is not an IntervalProd instance'
                            ''.format(intv_prod))
        if not isinstance(grid, RectGrid):
            raise TypeError('{!r} is not a RectGrid instance'
                            ''.format(grid))

        # More conclusive error than the one from contains_set
        if intv_prod.ndim != grid.ndim:
            raise ValueError('interval product {} is {}-dimensional while '
                             'grid {} is {}-dimensional'
                             ''.format(intv_prod, intv_prod.ndim,
                                       grid, grid.ndim))

        if not intv_prod.contains_set(grid):
            raise ValueError('{} is not contained in {}'
                             ''.format(grid, intv_prod))

        self.__set = intv_prod
        self.__grid = grid

        # Initialize the cell boundaries, the defining property of partitions
        bdry_vecs = []
        for ax, vec in enumerate(self.grid.coord_vectors):
            bdry = np.empty(len(vec) + 1)
            bdry[1:-1] = (vec[1:] + vec[:-1]) / 2.0
            bdry[0] = self.min()[ax]
            bdry[-1] = self.max()[ax]
            bdry_vecs.append(bdry)

        self.__cell_boundary_vecs = tuple(bdry_vecs)

        # Initialize nodes_on_bdry
        left_on_bdry = np.isclose(self.grid.min_pt, self.set.min_pt)[:, None]
        right_on_bdry = np.isclose(self.grid.max_pt, self.set.max_pt)[:, None]
        on_bdry = np.hstack([left_on_bdry, right_on_bdry]).tolist()
        self.__nodes_on_bdry = tuple(tuple(r) for r in on_bdry)

    @property
    def cell_boundary_vecs(self):
        """Return the cell boundaries as coordinate vectors.

        Examples
        --------
        >>> rect = odl.IntervalProd([0, -1], [1, 2])
        >>> grid = odl.RectGrid([0, 1], [-1, 0, 2])
        >>> part = odl.RectPartition(rect, grid)
        >>> part.cell_boundary_vecs
        (array([ 0. ,  0.5,  1. ]), array([-1. , -0.5,  1. ,  2. ]))
        """
        return self.__cell_boundary_vecs

    @property
    def set(self):
        """Partitioned set, an `IntervalProd`."""
        return self.__set

    @property
    def nodes_on_bdry(self):
        """Encoding of grid points lying on the boundary.

        Examples
        --------
        Using global option (default ``False``):

        >>> part = odl.nonuniform_partition([0, 2, 3], [1, 3])
        >>> part.nodes_on_bdry
        False
        >>> part = odl.nonuniform_partition([0, 2, 3], [1, 3],
        ...                                 nodes_on_bdry=True)
        >>> part.nodes_on_bdry
        True

        ``False`` in axis 0, ``True`` in axis 1:

        >>> part = odl.nonuniform_partition([0, 2, 3], [1, 3],
        ...                                 nodes_on_bdry=[False, True])
        >>> part.nodes_on_bdry
        (False, True)

        In axis 0, ``False`` left and ``True`` right, in axis 1 ``False``:

        >>> part = odl.nonuniform_partition([0, 2, 3], [1, 3],
        ...                                 nodes_on_bdry=[[False, True],
        ...                                                False])
        >>> part.nodes_on_bdry
        ((False, True), False)
        """
        if self.size == 0:
            return True

        nodes_on_bdry = []
        for on_bdry in self.nodes_on_bdry_byaxis:
            left, right = on_bdry
            if left == right:
                nodes_on_bdry.append(left)
            else:
                nodes_on_bdry.append((left, right))
        if all(on_bdry == nodes_on_bdry[0] for on_bdry in nodes_on_bdry[1:]):
            return nodes_on_bdry[0]
        else:
            return tuple(nodes_on_bdry)

    @property
    def nodes_on_bdry_byaxis(self):
        """Nested tuple of booleans for `nodes_on_bdry`.

        This attribute is equivalent to `nodes_on_bdry`, but always in
        the form of a nested tuple.
        """
        return self.__nodes_on_bdry

    # IntervalProd related pass-through methods and derived properties
    # min, max and extent are for duck-typing purposes
    @property
    def min_pt(self):
        """Minimum coordinates of the partitioned set."""
        return self.set.min_pt

    @property
    def max_pt(self):
        """Maximum coordinates of the partitioned set."""
        return self.set.max_pt

    @property
    def mid_pt(self):
        """Midpoint of the partitioned set."""
        return self.set.mid_pt

    def min(self):
        """Return the minimum point of the partitioned set.

        See Also
        --------
        odl.core.set.domain.IntervalProd.min
        """
        return self.set.min()

    def max(self):
        """Return the maximum point of the partitioned set.

        See Also
        --------
        odl.core.set.domain.IntervalProd.max
        """
        return self.set.max()

    @property
    def extent(self):
        """Return a vector containing the total extent (max - min)."""
        return self.set.extent

    @property
    def grid(self):
        """`RectGrid` defining this partition."""
        return self.__grid

    # RectGrid related pass-through methods and derived properties
    @property
    def is_uniform_byaxis(self):
        """Boolean tuple showing uniformity of ``self.grid`` per axis.

        Examples
        --------
        >>> part = nonuniform_partition([0, 1, 3], [1, 2, 3])
        >>> part.is_uniform_byaxis
        (False, True)
        """
        return self.grid.is_uniform_byaxis

    @property
    def is_uniform(self):
        """``True`` if `grid` is uniform."""
        return self.grid.is_uniform

    @property
    def has_isotropic_cells(self):
        """``True`` if `grid` is uniform and `cell_sides` are all equal.

        Always ``True`` for 1D partitions.

        Examples
        --------
        >>> part = uniform_partition([0, -1], [1, 1], (5, 10))
        >>> part.has_isotropic_cells
        True
        >>> part = uniform_partition([0, -1], [1, 1], (5, 5))
        >>> part.has_isotropic_cells
        False
        """
        return self.is_uniform and np.allclose(self.cell_sides[:-1],
                                               self.cell_sides[1:])

    @property
    def ndim(self):
        """Number of dimensions."""
        return self.grid.ndim

    @property
    def shape(self):
        """Number of cells per axis, equal to ``self.grid.shape``."""
        return self.grid.shape

    @property
    def size(self):
        """Total number of cells, equal to ``self.grid.size``."""
        return self.grid.size

    def __len__(self):
        """Return ``len(self)``.

        Total number of cells along the first dimension.

        Examples
        --------
        >>> partition = odl.uniform_partition([0, 0, 0],
        ...                                   [1, 1, 1],
        ...                                   shape=(2, 3, 4))
        >>> len(partition)
        2

        See Also
        --------
        size : The total number of cells.
        """
        return len(self.grid)

    def points(self, order='C'):
        """Return the sampling grid points.

        See Also
        --------
        odl.core.discr.grid.RectGrid.points
        """
        return self.grid.points(order)

    @property
    def meshgrid(self):
        """Return the sparse meshgrid of sampling points."""
        return self.grid.meshgrid

    @property
    def coord_vectors(self):
        """Coordinate vectors of the grid."""
        return self.grid.coord_vectors

    # Further derived methods / properties
    @property
    def boundary_cell_fractions(self):
        """Return a tuple of contained fractions of boundary cells.

        Since the outermost grid points can have any distance to the
        boundary of the partitioned set, the "natural" outermost cell
        around these points can either be cropped or extended. This
        property is a tuple of (float, float) tuples, one entry per
        dimension, where the fractions of the left- and rightmost
        cells inside the set are stored. If a grid point lies exactly
        on the boundary, the value is 1/2 since the cell is cut in half.
        Otherwise, any value larger than 1/2 is possible.

        Returns
        -------
        on_bdry : tuple of 2-tuples of floats
            Each 2-tuple contains the fraction of the leftmost
            (first entry) and rightmost (second entry) cell in the
            partitioned set in the corresponding dimension.

        See Also
        --------
        cell_boundary_vecs

        Examples
        --------
        We create a partition of the rectangle [0, 1.5] x [-2, 2] with
        the grid points [0, 1] x [-1, 0, 2]. The "natural" cells at the
        boundary would be:

            [-0.5, 0.5] and [0.5, 1.5] in the first axis

            [-1.5, -0.5] and [1, 3] in the second axis

        Thus, in the first axis, the fractions contained in [0, 1.5]
        are 0.5 and 1, and in the second axis, [-2, 2] contains the
        fractions 1.5 and 0.5.

        >>> rect = odl.IntervalProd([0, -2], [1.5, 2])
        >>> grid = odl.RectGrid([0, 1], [-1, 0, 2])
        >>> part = odl.RectPartition(rect, grid)
        >>> part.boundary_cell_fractions
        ((0.5, 1.0), (1.5, 0.5))
        """
        frac_list = []
        for ax, (cvec, bmin, bmax) in enumerate(zip(
                self.grid.coord_vectors, self.set.min_pt, self.set.max_pt)):
            # Degenerate axes have a value of 1.0 (this is used as weight
            # in integration formulas later)
            if len(cvec) == 1:
                frac_list.append((1.0, 1.0))
            else:
                left_frac = 0.5 + (cvec[0] - bmin) / (cvec[1] - cvec[0])
                right_frac = 0.5 + (bmax - cvec[-1]) / (cvec[-1] - cvec[-2])
                frac_list.append((left_frac, right_frac))

        return tuple(frac_list)

    @property
    def cell_sizes_vecs(self):
        """Return the cell sizes as coordinate vectors.

        Returns
        -------
        csizes : tuple of `numpy.ndarray`'s
            The cell sizes per axis. The length of the vectors is the
            same as the corresponding ``grid.coord_vectors``.
            For axes with 1 grid point, cell size is set to 0.0.

        Examples
        --------
        We create a partition of the rectangle [0, 1] x [-1, 2] into
        2 x 3 cells with the grid points [0, 1] x [-1, 0, 2]. This
        implies that the cell boundaries are given as
        [0, 0.5, 1] x [-1, -0.5, 1, 2], hence the cell size vectors
        are [0.5, 0.5] x [0.5, 1.5, 1]:

        >>> rect = odl.IntervalProd([0, -1], [1, 2])
        >>> grid = odl.RectGrid([0, 1], [-1, 0, 2])
        >>> part = odl.RectPartition(rect, grid)
        >>> part.cell_boundary_vecs
        (array([ 0. ,  0.5,  1. ]), array([-1. , -0.5,  1. ,  2. ]))
        >>> part.cell_sizes_vecs
        (array([ 0.5,  0.5]), array([ 0.5,  1.5,  1. ]))
        """
        csizes = []
        for ax, cvec in enumerate(self.grid.coord_vectors):
            if len(cvec) == 1:
                csizes.append(np.array([0.0]))
            else:
                csize = np.empty_like(cvec)
                csize[1:-1] = (cvec[2:] - cvec[:-2]) / 2.0
                csize[0] = (cvec[0] + cvec[1]) / 2 - self.min()[ax]
                csize[-1] = self.max()[ax] - (cvec[-2] + cvec[-1]) / 2
                csizes.append(csize)

        return tuple(csizes)

    @property
    def cell_sides(self):
        """Side lengths of all 'inner' cells of a uniform partition.

        Only defined if ``self.grid`` is uniform.

        Examples
        --------
        We create a partition of the rectangle [0, 1] x [-1, 2] into
        3 x 3 cells, where the grid points lie on the boundary. This
        means that the grid points are [0, 0.5, 1] x [-1, 0.5, 2],
        i.e. the inner cell has side lengths 0.5 x 1.5:

        >>> rect = odl.IntervalProd([0, -1], [1, 2])
        >>> grid = odl.uniform_grid([0, -1], [1, 2], (3, 3))
        >>> part = odl.RectPartition(rect, grid)
        >>> part.cell_sides
        array([ 0.5,  1.5])
        """
        sides = self.grid.stride
        sides[sides == 0] = self.extent[sides == 0]
        return sides

    @property
    def cell_volume(self):
        """Volume of the 'inner' cells of a uniform partition.

        Only defined if ``self.grid`` is uniform.

        Examples
        --------
        We create a partition of the rectangle [0, 1] x [-1, 2] into
        3 x 3 cells, where the grid points lie on the boundary. This
        means that the grid points are [0, 0.5, 1] x [-1, 0.5, 2],
        i.e. the inner cell has side lengths 0.5 x 1.5:

        >>> rect = odl.IntervalProd([0, -1], [1, 2])
        >>> grid = odl.uniform_grid([0, -1], [1, 2], (3, 3))
        >>> part = odl.RectPartition(rect, grid)
        >>> part.cell_sides
        array([ 0.5,  1.5])
        >>> part.cell_volume
        0.75
        """
        return 0.0 if self.size == 0 else float(np.prod(self.cell_sides))

    def approx_equals(self, other, atol):
        """Return ``True`` in case of approximate equality.

        Returns
        -------
        approx_eq : bool
            ``True`` if ``other`` is a `RectPartition` instance with
            ``self.set == other.set`` up to ``atol`` and
            ``self.grid == other.other`` up to ``atol``, ``False`` otherwise.
        """
        if other is self:
            return True
        elif not isinstance(other, RectPartition):
            return False
        else:
            return (self.set.approx_equals(other.set, atol=atol) and
                    self.grid.approx_equals(other.grid, atol=atol))

    def __eq__(self, other):
        """Return ``self == other``."""
        # Implemented separately for performance reasons
        if other is self:
            return True

        # Optimized version for exact equality
        return (type(other) is type(self) and
                self.set == other.set and
                self.grid == other.grid)

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash((type(self), self.set, self.grid))

    def __ne__(self, other):
        """Return ``self != other``."""
        return not (self == other)

    def __getitem__(self, indices):
        """Return ``self[indices]``.

        Parameters
        ----------
        indices : index expression
            Object determining which parts of the partition to extract.
            ``None`` (new axis) and empty axes are not supported.

        Examples
        --------
        Take every second grid point. Note that is is in general non-uniform:

        >>> partition = odl.uniform_partition(0, 10, 10)
        >>> partition[::2]
        nonuniform_partition(
            [ 0.5,  2.5,  4.5,  6.5,  8.5],
            min_pt=0.0, max_pt=10.0
        )

        A more advanced example is:

        >>> intvp = odl.IntervalProd([-1, 1, 4, 2], [3, 6, 5, 7])
        >>> grid = odl.RectGrid([-1, 0, 3], [2, 4], [5], [2, 4, 7])
        >>> part = odl.RectPartition(intvp, grid)
        >>> part
        nonuniform_partition(
            [-1.,  0.,  3.],
            [ 2.,  4.],
            [ 5.],
            [ 2.,  4.,  7.],
            min_pt=[-1.,  1.,  4.,  2.], max_pt=[ 3.,  6.,  5.,  7.]
        )

        Take an advanced slice (every second along the first axis,
        the last in the last axis and everything in between):

        >>> part[::2, ..., -1]
        nonuniform_partition(
            [-1.,  3.],
            [ 2.,  4.],
            [ 5.],
            [ 7.],
            min_pt=[-1. ,  1. ,  4. ,  5.5], max_pt=[ 3.,  6.,  5.,  7.]
        )

        Too few indices are filled up with an ellipsis from the right:

        >>> part[1]
        nonuniform_partition(
            [ 0.],
            [ 2.,  4.],
            [ 5.],
            [ 2.,  4.,  7.],
            min_pt=[-0.5,  1. ,  4. ,  2. ], max_pt=[ 1.5,  6. ,  5. ,  7. ]
        )

        Colons etc work as expected:

        >>> part[:] == part
        True
        >>> part[:, :, :] == part
        True
        >>> part[...] == part
        True
        """
        # Special case of index list: slice along first axis
        if isinstance(indices, list):
            if indices:
                new_min_pt = [self.cell_boundary_vecs[0][:-1][indices][0]]
                new_max_pt = [self.cell_boundary_vecs[0][1:][indices][-1]]
                for cvec in self.cell_boundary_vecs[1:]:
                    new_min_pt.append(cvec[0])
                    new_max_pt.append(cvec[-1])
            else:
                new_min_pt = new_max_pt = []

            new_intvp = IntervalProd(new_min_pt, new_max_pt)
            new_grid = self.grid[indices]
            return RectPartition(new_intvp, new_grid)

        indices = normalized_index_expression(indices, self.shape,
                                              int_to_slice=True)
        # Build the new partition
        new_min_pt, new_max_pt = [], []
        for cvec, idx in zip(self.cell_boundary_vecs, indices):
            # Determine the subinterval min_pt and max_pt vectors. Take the
            # first min_pt as new min_pt and the last max_pt as new max_pt.
            if isinstance(idx, slice):
                # Only use the slice to extract min and max without using
                # the step size. This is in order for expressions like
                # self[::2] to not change the maximum.
                idx = slice(idx.start, idx.stop, None)

            sub_min_pt = cvec[:-1][idx]
            sub_max_pt = cvec[1:][idx]
            new_min_pt.append(sub_min_pt[0])
            new_max_pt.append(sub_max_pt[-1])

        new_intvp = IntervalProd(new_min_pt, new_max_pt)
        new_grid = self.grid[indices]
        return RectPartition(new_intvp, new_grid)

    def insert(self, index, *parts):
        """Return a copy with ``parts`` inserted before ``index``.

        The given partitions are inserted (as a block) into ``self``,
        yielding a new partition whose number of dimensions is the sum of
        the numbers of dimensions of all involved partitions.
        Note that no changes are made in-place.

        Parameters
        ----------
        index : int
            Index of the dimension before which ``other`` is to
            be inserted. Negative indices count backwards from
            ``self.ndim``.
        part1, ..., partN : `RectPartition`
            Partitions to be inserted into ``self``.

        Returns
        -------
        newpart : `RectPartition`
            The enlarged partition.

        Examples
        --------
        >>> part1 = odl.uniform_partition([0, -1], [1, 2], (3, 3))
        >>> part2 = odl.uniform_partition(0, 1, 5)
        >>> part1.insert(1, part2)
        uniform_partition([ 0.,  0., -1.], [ 1.,  1.,  2.], (3, 5, 3))

        See Also
        --------
        append
        """
        if not all(isinstance(p, RectPartition) for p in parts):
            raise TypeError('`parts` must all be `RectPartition` instances, '
                            'got ({})'
                            ''.format(', '.join(repr(p) for p in parts)))

        newgrid = self.grid.insert(index, *(p.grid for p in parts))
        newset = self.set.insert(index, *(p.set for p in parts))
        return RectPartition(newset, newgrid)

    def append(self, *parts):
        """Insert ``parts`` at the end as a block.

        Parameters
        ----------
        part1, ..., partN : `RectPartition`
            Partitions to be appended to ``self``.

        Returns
        -------
        newpart : `RectPartition`
            The enlarged partition.

        Examples
        --------
        >>> part1 = odl.uniform_partition(-1, 2, 3)
        >>> part2 = odl.uniform_partition(0, 1, 5)
        >>> part1.append(part2)
        uniform_partition([-1.,  0.], [ 2.,  1.], (3, 5))
        >>> part1.append(part2, part2)
        uniform_partition([-1.,  0.,  0.], [ 2.,  1.,  1.], (3, 5, 5))

        See Also
        --------
        insert
        """
        return self.insert(self.ndim, *parts)

    def squeeze(self, axis=None):
        """Return the partition with removed degenerate (length 1) dimensions.

        Parameters
        ----------
        axis : None or index expression, optional
            Subset of the axes to squeeze. Default: All axes.

        Returns
        -------
        squeezed : `RectPartition`
            Squeezed partition.

        Examples
        --------
        >>> p = odl.uniform_partition([0, -1], [1, 2], (3, 1))
        >>> p.squeeze()
        uniform_partition(0.0, 1.0, 3)

        The axis argument can be used to only squeeze some axes (if applicable)

        >>> p.squeeze(axis=0)
        uniform_partition([ 0., -1.], [ 1.,  2.], (3, 1))

        Notes
        -----
        This is not equivalent to
        ``RectPartiton(self.set.squeeze(), self.grid.squeeze())`` since the
        definition of degenerate is different in sets and grids. This method
        follow the definition used in grids, that is, an axis is degenerate if
        it has only one element.

        See Also
        --------
        odl.core.discr.grid.RectGrid.squeeze
        odl.core.set.domain.IntervalProd.squeeze
        """
        if axis is None:
            rng = range(self.ndim)
        else:
            rng = list(np.atleast_1d(np.arange(self.ndim)[axis]))

        new_indcs = [i for i in range(self.ndim)
                     if i not in rng or self.grid.nondegen_byaxis[i]]
        newset = self.set[new_indcs]
        return RectPartition(newset, self.grid.squeeze(axis))

    def index(self, value, floating=False):
        """Return the index of a value in the domain.

        Parameters
        ----------
        value : ``self.set`` element
            Point whose index to find.
        floating : bool, optional
            If True, then the index should also give the position inside the
            voxel. This is given by returning the integer valued index of the
            voxel plus the distance from the left cell boundary as a fraction
            of the full cell size.

        Returns
        -------
        index : int, float, tuple of int or tuple of float
            Index of the value, as counted from the left.
            If ``self.ndim > 1`` the result is a tuple, else a scalar.
            If ``floating=True`` the scalar is a float, else an int.

        Examples
        --------
        Get the indices of start and end:

        >>> p = odl.uniform_partition(0, 2, 5)
        >>> p.index(0)
        0
        >>> p.index(2)
        4

        For points inside voxels, the index of the containing cell is returned:

        >>> p.index(0.2)
        0

        By using the ``floating`` argument, partial positions inside the voxels
        can instead be determined:

        >>> p.index(0.2, floating=True)
        0.5

        These indices work with indexing, extracting the voxel in which the
        point lies:

        >>> p[p.index(0.1)]
        uniform_partition(0.0, 0.4, 1)

        The same principle also works in higher dimensions:

        >>> p = uniform_partition([0, -1], [1, 2], (4, 1))
        >>> p.index([0.5, 2])
        (2, 0)
        >>> p[p.index([0.5, 2])]
        uniform_partition([ 0.5, -1. ], [ 0.75,  2.  ], (1, 1))
        """
        value = np.atleast_1d(self.set.element(value))
        result = []
        for val, cell_bdry_vec in zip(value, self.cell_boundary_vecs):
            ind = np.searchsorted(cell_bdry_vec, val)
            if floating:
                if cell_bdry_vec[ind] == val:
                    # Value is on top of edge
                    result.append(float(ind))
                else:
                    # interpolate between
                    csize = float(cell_bdry_vec[ind] - cell_bdry_vec[ind - 1])
                    result.append(ind - (cell_bdry_vec[ind] - val) / csize)
            else:
                if cell_bdry_vec[ind] == val and ind != len(cell_bdry_vec) - 1:
                    # Value is on top of edge, but not last edge
                    result.append(ind)
                else:
                    result.append(ind - 1)

        if self.ndim == 1:
            result = result[0]
        else:
            result = tuple(result)

        return result

    @property
    def byaxis(self):
        """Object to index ``self`` along axes.

        Examples
        --------
        Indexing with integers or slices:

        >>> p = odl.uniform_partition([0, 1, 2], [1, 3, 5], (3, 5, 6))
        >>> p.byaxis[0]
        uniform_partition(0.0, 1.0, 3)
        >>> p.byaxis[1]
        uniform_partition(1.0, 3.0, 5)
        >>> p.byaxis[2]
        uniform_partition(2.0, 5.0, 6)
        >>> p.byaxis[:] == p
        True
        >>> p.byaxis[1:]
        uniform_partition([ 1.,  2.], [ 3.,  5.], (5, 6))

        Lists can be used to stack subpartitions arbitrarily:

        >>> p.byaxis[[0, 2, 0]]
        uniform_partition([ 0.,  2.,  0.], [ 1.,  5.,  1.], (3, 6, 3))
        """
        partition = self

        class RectPartitionByAxis(object):

            """Helper class for accessing `RectPartition` by axis."""

            def __getitem__(self, indices):
                """Return ``self[indices]``."""
                try:
                    iter(indices)
                except TypeError:
                    # Slice or integer
                    slc = np.zeros(partition.ndim, dtype=object)
                    slc[indices] = slice(None)
                    squeeze_axes = np.where(slc == 0)[0]
                    newpart = partition[tuple(slc)].squeeze(squeeze_axes)
                else:
                    # Sequence, stack together from single-integer indexing
                    indices = [int(i) for i in indices]
                    byaxis = partition.byaxis
                    parts = [byaxis[i] for i in indices]
                    if not parts:
                        newpart = uniform_partition([], [], ())
                    else:
                        newpart = parts[0].append(*(parts[1:]))

                return newpart

            def __repr__(self):
                """Return ``repr(self)``.

                Examples
                --------
                >>> p = odl.uniform_partition(0, 1, 5)
                >>> p.byaxis
                uniform_partition(0, 1, 5).byaxis
                """
                return '{!r}.byaxis'.format(partition)

        return RectPartitionByAxis()

    def __repr__(self):
        """Return ``repr(self)``."""
        if self.ndim == 0:
            return 'uniform_partition([], [], ())'

        bdry_fracs = np.vstack(self.boundary_cell_fractions)
        default_bdry_fracs = np.all(np.isclose(bdry_fracs, 0.5) |
                                    np.isclose(bdry_fracs, 1.0))

        # Get default shifts of min_pt and max_pt from corresponding
        # grid points
        csizes_l = np.fromiter((s[0] for s in self.cell_sizes_vecs),
                               dtype=float)
        csizes_r = np.fromiter((s[-1] for s in self.cell_sizes_vecs),
                               dtype=float)

        shift_l = ((bdry_fracs[:, 0].astype(float).squeeze() - 0.5) *
                   csizes_l)
        shift_r = ((bdry_fracs[:, 1].astype(float).squeeze() - 0.5) *
                   csizes_r)

        if self.is_uniform and default_bdry_fracs:
            ctor = 'uniform_partition'
            if self.ndim == 1:
                posargs = [self.min_pt[0], self.max_pt[0], self.shape[0]]
                posmod = [':.4', ':.4', '']
            else:
                posargs = [self.min_pt, self.max_pt, self.shape]
                posmod = [array_str, array_str, '']

            optargs = [('nodes_on_bdry', self.nodes_on_bdry, False)]

            with npy_printoptions(precision=4):
                sig_str = signature_string(posargs, optargs, mod=[posmod, ''])
            return '{}({})'.format(ctor, sig_str)
        else:
            ctor = 'nonuniform_partition'
            posargs = self.coord_vectors
            posmod = array_str

            optargs = []
            # Defaults with and without nodes_on_bdry option
            nodes_def_min_pt = self.grid.min_pt - shift_l
            nodes_def_max_pt = self.grid.max_pt + shift_r
            def_min_pt = self.grid.min_pt - 0.5 * csizes_l
            def_max_pt = self.grid.max_pt + 0.5 * csizes_r

            # Since min/max_pt and nodes_on_bdry are mutex, we need a
            # couple of cases here
            optmod = []
            if (np.allclose(self.min_pt, nodes_def_min_pt) and
                    np.allclose(self.max_pt, nodes_def_max_pt)):
                # Append nodes_on_bdry to list of optional args
                optargs.append(('nodes_on_bdry', self.nodes_on_bdry, False))
                optmod.append('')
            else:
                # Append min/max_pt to list of optional args if not
                # default (need check manually because array comparison is
                # ambiguous)
                if not np.allclose(self.min_pt, def_min_pt):
                    if self.ndim == 1:
                        optargs.append(('min_pt', self.min_pt[0], None))
                        optmod.append(':.4')
                    else:
                        with npy_printoptions(precision=4):
                            optargs.append(
                                ('min_pt', array_str(self.min_pt), ''))
                        optmod.append('!s')

                if not np.allclose(self.max_pt, def_max_pt):
                    if self.ndim == 1:
                        optargs.append(('max_pt', self.max_pt[0], None))
                        optmod.append(':.4')
                    else:
                        with npy_printoptions(precision=4):
                            optargs.append(
                                ('max_pt', array_str(self.max_pt), ''))
                        optmod.append('!s')

            sig_str = signature_string(posargs, optargs, mod=[posmod, optmod],
                                       sep=[',\n', ', ', ',\n'])
            return '{}(\n{}\n)'.format(ctor, indent(sig_str))

    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)


def uniform_partition_fromintv(intv_prod, shape, nodes_on_bdry=False):
    """Return a partition of an interval product into equally sized cells.

    Parameters
    ----------
    intv_prod : `IntervalProd`
        Interval product to be partitioned
    shape : int or sequence of ints
        Number of nodes per axis. For 1d intervals, a single integer
        can be specified.
    nodes_on_bdry : bool or sequence, optional
        If a sequence is provided, it determines per axis whether to
        place the last grid point on the boundary (``True``) or shift it
        by half a cell size into the interior (``False``). In each axis,
        an entry may consist in a single bool or a 2-tuple of
        bool. In the latter case, the first tuple entry decides for
        the left, the second for the right boundary. The length of the
        sequence must be ``intv_prod.ndim``.

        A single boolean is interpreted as a global choice for all
        boundaries.

    See Also
    --------
    uniform_partition_fromgrid

    Examples
    --------
    By default, no grid points are placed on the boundary:

    >>> interval = odl.IntervalProd(0, 1)
    >>> part = odl.uniform_partition_fromintv(interval, 4)
    >>> part.cell_boundary_vecs
    (array([ 0.  ,  0.25,  0.5 ,  0.75,  1.  ]),)
    >>> part.grid.coord_vectors
    (array([ 0.125,  0.375,  0.625,  0.875]),)

    This can be changed with the nodes_on_bdry parameter:

    >>> part = odl.uniform_partition_fromintv(interval, 3,
    ...                                       nodes_on_bdry=True)
    >>> part.cell_boundary_vecs
    (array([ 0.  ,  0.25,  0.75,  1.  ]),)
    >>> part.grid.coord_vectors
    (array([ 0. ,  0.5,  1. ]),)

    We can specify this per axis, too. In this case we choose both
    in the first axis and only the rightmost in the second:

    >>> rect = odl.IntervalProd([0, 0], [1, 1])
    >>> part = odl.uniform_partition_fromintv(
    ...     rect, (3, 3), nodes_on_bdry=(True, (False, True)))
    ...
    >>> part.cell_boundary_vecs[0]  # first axis, as above
    array([ 0.  ,  0.25,  0.75,  1.  ])
    >>> part.grid.coord_vectors[0]
    array([ 0. ,  0.5,  1. ])
    >>> part.cell_boundary_vecs[1]  # second, asymmetric axis
    array([ 0. ,  0.4,  0.8,  1. ])
    >>> part.grid.coord_vectors[1]
    array([ 0.2,  0.6,  1. ])
    """
    grid = uniform_grid_fromintv(intv_prod, shape, nodes_on_bdry=nodes_on_bdry)
    return RectPartition(intv_prod, grid)


def uniform_partition_fromgrid(grid, min_pt=None, max_pt=None):
    """Return a partition of an interval product based on a given grid.

    This method is complementary to `uniform_partition_fromintv` in that
    it infers the set to be partitioned from a given grid and optional
    parameters for ``min_pt`` and ``max_pt`` of the set.

    Parameters
    ----------
    grid : `RectGrid`
        Grid on which the partition is based
    min_pt, max_pt : float, sequence of floats, or dict, optional
        Spatial points defining the lower/upper limits of the intervals
        to be partitioned. The points can be specified in two ways:

        float or sequence: The values are used directly as ``min_pt``
        and/or ``max_pt``.

        dict: Index-value pairs specifying an axis and a spatial
        coordinate to be used in that axis. In axes which are not a key
        in the dictionary, the coordinate for the vector is calculated
        as::

            min_pt = x[0] - (x[1] - x[0]) / 2
            max_pt = x[-1] + (x[-1] - x[-2]) / 2

        See ``Examples`` below.

        In general, ``min_pt`` may not be larger than ``grid.min_pt``,
        and ``max_pt`` not smaller than ``grid.max_pt`` in any component.
        ``None`` is equivalent to an empty dictionary, i.e. the values
        are calculated in each dimension.

    See Also
    --------
    uniform_partition_fromintv

    Examples
    --------
    Have ``min_pt`` and ``max_pt`` of the bounding box automatically
    calculated:

    >>> grid = odl.uniform_grid(0, 1, 3)
    >>> grid.coord_vectors
    (array([ 0. ,  0.5,  1. ]),)
    >>> part = odl.uniform_partition_fromgrid(grid)
    >>> part.cell_boundary_vecs
    (array([-0.25,  0.25,  0.75,  1.25]),)

    ``min_pt`` and ``max_pt`` can be given explicitly:

    >>> part = odl.uniform_partition_fromgrid(grid, min_pt=0, max_pt=1)
    >>> part.cell_boundary_vecs
    (array([ 0.  ,  0.25,  0.75,  1.  ]),)

    Using dictionaries, selective axes can be explicitly set. The
    keys refer to axes, the values to the coordinates to use:

    >>> grid = odl.uniform_grid([0, 0], [1, 1], (3, 3))
    >>> part = odl.uniform_partition_fromgrid(grid,
    ...                                       min_pt={0: -1}, max_pt={-1: 3})
    >>> part.cell_boundary_vecs[0]
    array([-1.  ,  0.25,  0.75,  1.25])
    >>> part.cell_boundary_vecs[1]
    array([-0.25,  0.25,  0.75,  3.  ])
    """
    # Make dictionaries from `min_pt` and `max_pt` and fill with `None` where
    # no value is given (taking negative indices into account)
    if min_pt is None:
        min_pt = {i: None for i in range(grid.ndim)}
    elif not hasattr(min_pt, 'items'):  # array-like
        min_pt = np.atleast_1d(min_pt)
        min_pt = {i: float(v) for i, v in enumerate(min_pt)}
    else:
        min_pt.update({i: None for i in range(grid.ndim)
                       if i not in min_pt and i - grid.ndim not in min_pt})

    if max_pt is None:
        max_pt = {i: None for i in range(grid.ndim)}
    elif not hasattr(max_pt, 'items'):
        max_pt = np.atleast_1d(max_pt)
        max_pt = {i: float(v) for i, v in enumerate(max_pt)}
    else:
        max_pt.update({i: None for i in range(grid.ndim)
                      if i not in max_pt and i - grid.ndim not in max_pt})

    # Set the values in the vectors by computing (None) or directly from the
    # given vectors (otherwise).
    min_pt_vec = np.empty(grid.ndim)
    for ax, xmin in min_pt.items():
        if xmin is None:
            cvec = grid.coord_vectors[ax]
            if len(cvec) == 1:
                raise ValueError('in axis {}: cannot calculate `min_pt` with '
                                 'only 1 grid point'.format(ax))
            min_pt_vec[ax] = cvec[0] - (cvec[1] - cvec[0]) / 2
        else:
            min_pt_vec[ax] = xmin

    max_pt_vec = np.empty(grid.ndim)
    for ax, xmax in max_pt.items():
        if xmax is None:
            cvec = grid.coord_vectors[ax]
            if len(cvec) == 1:
                raise ValueError('in axis {}: cannot calculate `max_pt` with '
                                 'only 1 grid point'.format(ax))
            max_pt_vec[ax] = cvec[-1] + (cvec[-1] - cvec[-2]) / 2
        else:
            max_pt_vec[ax] = xmax

    return RectPartition(IntervalProd(min_pt_vec, max_pt_vec), grid)


def uniform_partition(min_pt=None, max_pt=None, shape=None, cell_sides=None,
                      nodes_on_bdry=False):
    """Return a partition with equally sized cells.

    Parameters
    ----------
    min_pt, max_pt : float or sequence of float, optional
        Vectors defining the lower/upper limits of the intervals in an
        `IntervalProd` (a rectangular box). ``None`` entries mean
        "compute the value".
    shape : int or sequence of ints, optional
        Number of nodes per axis. ``None`` entries mean
        "compute the value".
    cell_sides : float or sequence of floats, optional
        Side length of the partition cells per axis. ``None`` entries mean
        "compute the value".
    nodes_on_bdry : bool or sequence, optional
        If a sequence is provided, it determines per axis whether to
        place the last grid point on the boundary (``True``) or shift it
        by half a cell size into the interior (``False``). In each axis,
        an entry may consist in a single bool or a 2-tuple of
        bool. In the latter case, the first tuple entry decides for
        the left, the second for the right boundary. The length of the
        sequence must be ``array.ndim``.

        A single boolean is interpreted as a global choice for all
        boundaries.

    Notes
    -----
    In each axis, 3 of the 4 possible parameters ``min_pt``, ``max_pt``,
    ``shape`` and ``cell_sides`` must be given. If all four are
    provided, they are checked for consistency.

    See Also
    --------
    uniform_partition_fromintv : partition an existing set
    uniform_partition_fromgrid : use an existing grid as basis

    Examples
    --------
    Any combination of three of the four parameters can be used for
    creation of a partition:

    >>> part = odl.uniform_partition(min_pt=0, max_pt=2, shape=4)
    >>> part.cell_boundary_vecs
    (array([ 0. ,  0.5,  1. ,  1.5,  2. ]),)
    >>> part = odl.uniform_partition(min_pt=0, shape=4, cell_sides=0.5)
    >>> part.cell_boundary_vecs
    (array([ 0. ,  0.5,  1. ,  1.5,  2. ]),)
    >>> part = odl.uniform_partition(max_pt=2, shape=4, cell_sides=0.5)
    >>> part.cell_boundary_vecs
    (array([ 0. ,  0.5,  1. ,  1.5,  2. ]),)
    >>> part = odl.uniform_partition(min_pt=0, max_pt=2, cell_sides=0.5)
    >>> part.cell_boundary_vecs
    (array([ 0. ,  0.5,  1. ,  1.5,  2. ]),)

    In higher dimensions, the parameters can be given differently in
    each axis. Where ``None`` is given, the value will be computed:

    >>> part = odl.uniform_partition(min_pt=[0, 0], max_pt=[1, 2],
    ...                              shape=[4, 2])
    >>> part.cell_boundary_vecs
    (array([ 0.  ,  0.25,  0.5 ,  0.75,  1.  ]), array([ 0.,  1.,  2.]))
    >>> part = odl.uniform_partition(min_pt=[0, 0], max_pt=[1, 2],
    ...                              shape=[None, 2], cell_sides=[0.25, None])
    >>> part.cell_boundary_vecs
    (array([ 0.  ,  0.25,  0.5 ,  0.75,  1.  ]), array([ 0.,  1.,  2.]))
    >>> part = odl.uniform_partition(min_pt=[0, None], max_pt=[None, 2],
    ...                              shape=[4, 2], cell_sides=[0.25, 1])
    >>> part.cell_boundary_vecs
    (array([ 0.  ,  0.25,  0.5 ,  0.75,  1.  ]), array([ 0.,  1.,  2.]))

    By default, no grid points are placed on the boundary:

    >>> part = odl.uniform_partition(0, 1, 4)
    >>> part.nodes_on_bdry
    False
    >>> part.cell_boundary_vecs
    (array([ 0.  ,  0.25,  0.5 ,  0.75,  1.  ]),)
    >>> part.grid.coord_vectors
    (array([ 0.125,  0.375,  0.625,  0.875]),)

    This can be changed with the nodes_on_bdry parameter:

    >>> part = odl.uniform_partition(0, 1, 3, nodes_on_bdry=True)
    >>> part.nodes_on_bdry
    True
    >>> part.cell_boundary_vecs
    (array([ 0.  ,  0.25,  0.75,  1.  ]),)
    >>> part.grid.coord_vectors
    (array([ 0. ,  0.5,  1. ]),)

    We can specify this per axis, too. In this case we choose both
    in the first axis and only the rightmost in the second:

    >>> part = odl.uniform_partition([0, 0], [1, 1], (3, 3),
    ...                              nodes_on_bdry=(True, (False, True)))
    ...
    >>> part.cell_boundary_vecs[0]  # first axis, as above
    array([ 0.  ,  0.25,  0.75,  1.  ])
    >>> part.grid.coord_vectors[0]
    array([ 0. ,  0.5,  1. ])
    >>> part.cell_boundary_vecs[1]  # second, asymmetric axis
    array([ 0. ,  0.4,  0.8,  1. ])
    >>> part.grid.coord_vectors[1]
    array([ 0.2,  0.6,  1. ])
    """
    # Normalize partition parameters

    # np.size(None) == 1, so that would screw it for sizes 0 of the rest
    sizes = [np.size(p) for p in (min_pt, max_pt, shape, cell_sides)
             if p is not None]
    ndim = int(np.max(sizes))

    min_pt = normalized_scalar_param_list(min_pt, ndim, param_conv=float,
                                          keep_none=True)
    max_pt = normalized_scalar_param_list(max_pt, ndim, param_conv=float,
                                          keep_none=True)
    shape = normalized_scalar_param_list(shape, ndim, param_conv=safe_int_conv,
                                         keep_none=True)
    cell_sides = normalized_scalar_param_list(cell_sides, ndim,
                                              param_conv=float, keep_none=True)

    nodes_on_bdry = normalized_nodes_on_bdry(nodes_on_bdry, ndim)

    # Calculate the missing parameters in min_pt, max_pt, shape
    for i, (xmin, xmax, n, dx, on_bdry) in enumerate(
            zip(min_pt, max_pt, shape, cell_sides, nodes_on_bdry)):
        num_params = sum(p is not None for p in (xmin, xmax, n, dx))
        if num_params < 3:
            raise ValueError('in axis {}: expected at least 3 of the '
                             'parameters `min_pt`, `max_pt`, `shape`, '
                             '`cell_sides`, got {}'
                             ''.format(i, num_params))

        # Unpack the tuple if possible, else use bool globally for this axis
        try:
            bdry_l, bdry_r = on_bdry
        except TypeError:
            bdry_l = bdry_r = on_bdry

        # For each node on the boundary, we subtract 1/2 from the number of
        # full cells between min_pt and max_pt.
        if xmin is None:
            min_pt[i] = xmax - (n - sum([bdry_l, bdry_r]) / 2.0) * dx
        elif xmax is None:
            max_pt[i] = xmin + (n - sum([bdry_l, bdry_r]) / 2.0) * dx
        elif n is None:
            # Here we add to n since (e-b)/s gives the reduced number of cells.
            n_calc = (xmax - xmin) / dx + sum([bdry_l, bdry_r]) / 2.0
            n_round = int(round(n_calc))
            if abs(n_calc - n_round) > 1e-5:
                raise ValueError('in axis {}: calculated number of nodes '
                                 '{} = ({} - {}) / {} too far from integer'
                                 ''.format(i, n_calc, xmax, xmin, dx))
            shape[i] = n_round
        elif dx is None:
            pass
        else:
            xmax_calc = xmin + (n - sum([bdry_l, bdry_r]) / 2.0) * dx
            if not np.isclose(xmax, xmax_calc):
                raise ValueError('in axis {}: calculated endpoint '
                                 '{} = {} + {} * {} too far from given '
                                 'endpoint {}.'
                                 ''.format(i, xmax_calc, xmin, n, dx, xmax))

    return uniform_partition_fromintv(
        IntervalProd(min_pt, max_pt), shape, nodes_on_bdry)


def nonuniform_partition(*coord_vecs, **kwargs):
    """Return a partition with un-equally sized cells.

    Parameters
    ----------
    coord_vecs1, ... coord_vecsN : `array-like`
        Arrays of coordinates of the mid-points of the partition cells.
    min_pt, max_pt : float or sequence of floats, optional
        Vectors defining the lower/upper limits of the intervals in an
        `IntervalProd` (a rectangular box). ``None`` entries mean
        "compute the value".
    nodes_on_bdry : bool or sequence, optional
        If a sequence is provided, it determines per axis whether to
        place the last grid point on the boundary (``True``) or shift it
        by half a cell size into the interior (``False``). In each axis,
        an entry may consist in a single bool or a 2-tuple of
        bool. In the latter case, the first tuple entry decides for
        the left, the second for the right boundary. The length of the
        sequence must be ``array.ndim``.

        A single boolean is interpreted as a global choice for all
        boundaries.

        Cannot be given with both min_pt and max_pt since they determine the
        same thing.

        Default: ``False``

    See Also
    --------
    uniform_partition : uniformly spaced points
    uniform_partition_fromintv : partition an existing set
    uniform_partition_fromgrid : use an existing grid as basis

    Examples
    --------
    With uniformly spaced points the result is the same as a
    uniform partition:

    >>> odl.nonuniform_partition([0, 1, 2, 3])
    uniform_partition(-0.5, 3.5, 4)
    >>> odl.nonuniform_partition([0, 1, 2, 3], [1, 2])
    uniform_partition([-0.5,  0.5], [ 3.5,  2.5], (4, 2))

    If the points are not uniformly spaced, a nonuniform partition is
    created. Note that the containing interval is calculated by assuming
    that the points are in the middle of the sub-intervals:

    >>> odl.nonuniform_partition([0, 1, 3])
    nonuniform_partition(
        [ 0.,  1.,  3.]
    )

    Higher dimensional partitions are created by specifying the gridpoints
    along each dimension:

    >>> odl.nonuniform_partition([0, 1, 3], [1, 2])
    nonuniform_partition(
        [ 0.,  1.,  3.],
        [ 1.,  2.]
    )

    Partitions with a single element are by default degenerate

    >>> odl.nonuniform_partition(1)
    uniform_partition(1.0, 1.0, 1, nodes_on_bdry=True)

    If the endpoints should be on the boundary, the ``nodes_on_bdry`` parameter
    can be used:

    >>> odl.nonuniform_partition([0, 1, 3], nodes_on_bdry=True)
    nonuniform_partition(
        [ 0.,  1.,  3.],
        nodes_on_bdry=True
    )

    Users can also manually specify the containing intervals dimensions by
    using the ``min_pt`` and ``max_pt`` arguments:

    >>> odl.nonuniform_partition([0, 1, 3], min_pt=-2, max_pt=3)
    nonuniform_partition(
        [ 0.,  1.,  3.],
        min_pt=-2.0, max_pt=3.0
    )
    """
    min_pt = kwargs.pop('min_pt', None)
    max_pt = kwargs.pop('max_pt', None)
    nodes_on_bdry = kwargs.pop('nodes_on_bdry', False)
    if kwargs:
        raise TypeError('unexpected keyword arguments: {}'.format(kwargs))

    # np.size(None) == 1
    sizes = [len(coord_vecs)] + [np.size(p) for p in (min_pt, max_pt)]
    ndim = int(np.max(sizes))

    min_pt = normalized_scalar_param_list(min_pt, ndim, param_conv=float,
                                          keep_none=True)
    max_pt = normalized_scalar_param_list(max_pt, ndim, param_conv=float,
                                          keep_none=True)
    nodes_on_bdry = normalized_nodes_on_bdry(nodes_on_bdry, ndim)

    # Calculate the missing parameters in min_pt, max_pt
    for i, (xmin, xmax, (bdry_l, bdry_r), coords) in enumerate(
            zip(min_pt, max_pt, nodes_on_bdry, coord_vecs)):
        # Check input for redundancy
        if xmin is not None and bdry_l:
            raise ValueError('in axis {}: got both `min_pt` and '
                             '`nodes_on_bdry=True`'.format(i))
        if xmax is not None and bdry_r:
            raise ValueError('in axis {}: got both `max_pt` and '
                             '`nodes_on_bdry=True`'.format(i))

        # Handle length 1 inputs
        coords = np.array(coords, copy=AVOID_UNNECESSARY_COPY, ndmin=1)

        # Compute boundary position if not given by user
        if xmin is None:
            if bdry_l or len(coords) == 1:
                min_pt[i] = coords[0]
            else:
                min_pt[i] = coords[0] - (coords[1] - coords[0]) / 2.0
        if xmax is None:
            if bdry_r or len(coords) == 1:
                max_pt[i] = coords[-1]
            else:
                max_pt[i] = coords[-1] + (coords[-1] - coords[-2]) / 2.0

    interval = IntervalProd(min_pt, max_pt)
    grid = RectGrid(*coord_vecs)
    return RectPartition(interval, grid)


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test for the Primal-Dual Hybrid Gradient algorithm."""

from __future__ import division
import numpy as np

import odl
from odl.solvers import pdhg
from odl.core.util.testutils import all_almost_equal

# Places for the accepted error when comparing results
PLACES = 8

# Algorithm parameters
TAU = 0.3
SIGMA = 0.7
THETA = 0.9

# Test data array
DATA = np.arange(6)


def test_pdhg_simple_space():
    """Test for the Primal-Dual Hybrid Gradient algorithm."""

    # Create a discretized image space
    space = odl.uniform_discr(0, 1, DATA.size)

    # Operator
    op = odl.IdentityOperator(space)

    # Starting point (image)
    discr_vec = op.domain.element(DATA)

    # Relaxation variable required to resume iteration
    discr_vec_relax = discr_vec.copy()

    # Dual variable required to resume iteration
    discr_dual = op.range.zero()

    # Functional, use the same functional for F^* and G
    f = odl.functionals.ZeroFunctional(space)
    g = f.convex_conj

    # Run the algorithm
    pdhg(discr_vec, f, g, op, niter=1, tau=TAU, sigma=SIGMA, theta=THETA,
         callback=None, x_relax=discr_vec_relax, y=discr_dual)

    # Explicit computation
    vec_expl = (1 - TAU * SIGMA) * DATA

    assert all_almost_equal(discr_vec, vec_expl, PLACES)

    # Explicit computation of the value of the relaxation variable
    vec_relax_expl = (1 + THETA) * vec_expl - THETA * DATA

    assert all_almost_equal(discr_vec_relax, vec_relax_expl, PLACES)

    # Resume iteration with previous x but without previous relaxation
    pdhg(discr_vec, f, g, op, niter=1, tau=TAU, sigma=SIGMA, theta=THETA)

    vec_expl *= (1 - SIGMA * TAU)
    assert all_almost_equal(discr_vec, vec_expl, PLACES)

    # Resume iteration with x1 as above and with relaxation parameter
    discr_vec[:] = vec_expl
    pdhg(discr_vec, f, g, op, niter=1, tau=TAU, sigma=SIGMA, theta=THETA,
         x_relax=discr_vec_relax, y=discr_dual)

    vec_expl = vec_expl - TAU * SIGMA * (DATA + vec_relax_expl)
    assert all_almost_equal(discr_vec, vec_expl, PLACES)

    # Test acceleration parameter: use output argument for the relaxation
    # variable since otherwise two iterations are required for the
    # relaxation to take effect on the input variable
    # Acceleration parameter gamma=0 corresponds to relaxation parameter
    # theta=1 without acceleration

    # Relaxation parameter 1 and no acceleration
    discr_vec = op.domain.element(DATA)
    discr_vec_relax_no_gamma = op.domain.element(DATA)
    pdhg(discr_vec, f, g, op, niter=1, tau=TAU, sigma=SIGMA, theta=1,
         gamma_primal=None, x_relax=discr_vec_relax_no_gamma)

    # Acceleration parameter 0, overwrites relaxation parameter
    discr_vec = op.domain.element(DATA)
    discr_vec_relax_g0 = op.domain.element(DATA)
    pdhg(discr_vec, f, g, op, niter=1, tau=TAU, sigma=SIGMA, theta=0,
         gamma_primal=0, x_relax=discr_vec_relax_g0)

    assert discr_vec != discr_vec_relax_no_gamma
    assert all_almost_equal(discr_vec_relax_no_gamma, discr_vec_relax_g0)

    # Test callback execution
    pdhg(discr_vec, f, g, op, niter=1, tau=TAU, sigma=SIGMA, theta=THETA,
         callback=odl.solvers.CallbackPrintIteration())


def test_pdhg_product_space():
    """Test the PDHG algorithm using a product space operator."""

    # Create a discretized image space
    space = odl.uniform_discr(0, 1, DATA.size)

    # Operator
    identity = odl.IdentityOperator(space)

    # Create broadcasting operator
    prod_op = odl.BroadcastOperator(identity, -2 * identity)

    # Starting point for explicit computation
    discr_vec_0 = prod_op.domain.element(DATA)

    # Copy to be overwritten by the algorithm
    discr_vec = discr_vec_0.copy()

    # Proximal operator using the same factory function for F^* and G
    f = odl.functionals.ZeroFunctional(prod_op.domain)
    g = odl.functionals.ZeroFunctional(prod_op.range).convex_conj

    # Run the algorithm
    pdhg(discr_vec, f, g, prod_op, niter=1, tau=TAU, sigma=SIGMA, theta=THETA)

    vec_expl = discr_vec_0 - TAU * SIGMA * prod_op.adjoint(
        prod_op(discr_vec_0))
    assert all_almost_equal(discr_vec, vec_expl, PLACES)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test ASTRA back-end using CUDA."""

from __future__ import division

import numpy as np
import pytest

import odl
from odl.applications.tomo.backends.astra_cuda import AstraCudaImpl
from odl.applications.tomo.util.testutils import skip_if_no_astra_cuda


# --- pytest fixtures --- #


# Find the valid projectors
projectors = [
    pytest.param(value, marks=skip_if_no_astra_cuda)
    for value in ['par2d', 'cone2d', 'par3d', 'cone3d', 'helical']
]

space_and_geometry_ids = [
    " geom='{}' ".format(p.values[0]) for p in projectors
]


@pytest.fixture(scope="module", params=projectors, ids=space_and_geometry_ids)
def space_and_geometry(request, odl_impl_device_pairs):
    impl, device = odl_impl_device_pairs
    dtype = 'float32'
    geom = request.param

    apart = odl.uniform_partition(0, 2 * np.pi, 8)

    if geom == 'par2d':
        reco_space = odl.uniform_discr([-4, -5], [4, 5], (4, 5),
                                       dtype=dtype, impl=impl, device=device)
        dpart = odl.uniform_partition(-6, 6, 6)
        geom = odl.applications.tomo.Parallel2dGeometry(apart, dpart)
    elif geom == 'par3d':
        reco_space = odl.uniform_discr([-4, -5, -6], [4, 5, 6], (4, 5, 6),
                                       dtype=dtype, impl=impl, device=device)
        dpart = odl.uniform_partition([-7, -8], [7, 8], (7, 8))
        geom = odl.applications.tomo.Parallel3dAxisGeometry(apart, dpart)
    elif geom == 'cone2d':
        reco_space = odl.uniform_discr([-4, -5], [4, 5], (4, 5),
                                       dtype=dtype, impl=impl, device=device)
        dpart = odl.uniform_partition(-6, 6, 6)
        geom = odl.applications.tomo.FanBeamGeometry(apart, dpart, src_radius=100,
                                        det_radius=10)
    elif geom == 'cone3d':
        reco_space = odl.uniform_discr([-4, -5, -6], [4, 5, 6], (4, 5, 6),
                                       dtype=dtype, impl=impl, device=device)
        dpart = odl.uniform_partition([-7, -8], [7, 8], (7, 8))

        geom = odl.applications.tomo.ConeBeamGeometry(apart, dpart,
                                         src_radius=200, det_radius=100)
    elif geom == 'helical':
        reco_space = odl.uniform_discr([-4, -5, -6], [4, 5, 6], (4, 5, 6),
                                       dtype=dtype, impl=impl, device=device)

        # overwrite angle
        apart = odl.uniform_partition(0, 2 * 2 * np.pi, 18)
        dpart = odl.uniform_partition([-7, -8], [7, 8], (7, 8))
        geom = odl.applications.tomo.ConeBeamGeometry(apart, dpart, pitch=1.0,
                                         src_radius=200, det_radius=100)
    else:
        raise ValueError('geom not valid')

    return reco_space, geom


# --- CUDA projector tests --- #


def test_astra_cuda_projector(space_and_geometry):
    """Test ASTRA CUDA projector."""

    # Create reco space and a phantom
    vol_space, geom = space_and_geometry
    phantom = odl.core.phantom.cuboid(vol_space)

    # Make projection space
    proj_space = odl.uniform_discr_frompartition(
        geom.partition,
        dtype=vol_space.dtype_identifier, 
        impl=vol_space.impl,
        device=vol_space.device)

    # create RayTransform implementation
    astra_cuda = AstraCudaImpl(geom, vol_space, proj_space)

    out = astra_cuda.proj_space.zero()
    # Forward evaluation
    proj_data = astra_cuda.call_forward(phantom)
    assert proj_data in proj_space
    assert proj_data.norm() > 0
    assert odl.all(0 <= proj_data)

    # Backward evaluation
    backproj = astra_cuda.call_backward(proj_data)
    assert backproj in vol_space
    assert backproj.norm() > 0
    assert odl.all(0 <= backproj)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for `discr_utils`."""

from __future__ import division

from functools import partial

import numpy as np
import pytest

import odl
from odl.core.discr.discr_utils import (
    linear_interpolator, nearest_interpolator, per_axis_interpolator,
    point_collocation, sampling_function)
from odl.core.discr.grid import sparse_meshgrid
from odl.core.util.testutils import all_almost_equal, all_equal, simple_fixture

from odl.core.array_API_support import lookup_array_backend, get_array_and_backend

# --- Helper functions --- #


def _test_eq(x, y):
    """Test equality of x and y."""
    assert x == y
    assert not x != y
    assert hash(x) == hash(y)


def _test_neq(x, y):
    """Test non-equality of x and y."""
    assert x != y
    assert not x == y
    assert hash(x) != hash(y)


def _points(domain, num):
    """Helper to generate ``num`` points in ``domain``."""
    min_pt = domain.min_pt
    max_pt = domain.max_pt
    ndim = domain.ndim
    points = np.random.uniform(low=0, high=1, size=(ndim, num))
    for i in range(ndim):
        points[i, :] = min_pt[i] + (max_pt[i] - min_pt[i]) * points[i]
    return points


def _meshgrid(domain, shape):
    """Helper to generate a ``shape`` meshgrid of points in ``domain``."""
    min_pt = domain.min_pt
    max_pt = domain.max_pt
    ndim = domain.ndim
    coord_vecs = []
    for i in range(ndim):
        vec = np.random.uniform(low=min_pt[i], high=max_pt[i], size=shape[i])
        vec.sort()
        coord_vecs.append(vec)
    return sparse_meshgrid(*coord_vecs)


class FuncList(list):  # So we can set __name__
    pass


# --- pytest fixtures (general) --- #


out_dtype = simple_fixture(
    'out_dtype',
    ['float32', 'float64', 'complex64'],
    fmt=' {name} = {value!r} '
)
domain_ndim = simple_fixture('domain_ndim', [1, 2])


# --- pytest fixtures (scalar test functions) --- #


def func_nd_oop(x):
    return sum(x)


def func_nd_ip(x, out):
    out[:] = sum(x)


def func_nd_dual(x, out=None):
    if out is None:
        return sum(x)
    else:
        out[:] = sum(x)


def func_nd_bcast_ref(x):
    return x[0] + 0 * sum(x[1:])


def func_nd_bcast_oop(x):
    return x[0]


def func_nd_bcast_ip(x, out):
    out[:] = x[0]


def func_nd_bcast_dual(x, out=None):
    if out is None:
        return x[0]
    else:
        out[:] = x[0]


func_nd_ref = func_nd_oop
# func_nd_params = [(func_nd_ref, f)
#                   for f in [func_nd_oop, func_nd_ip, func_nd_dual]]
# func_nd_params.extend([(func_nd_bcast_ref, func_nd_bcast_oop),
#                        (func_nd_bcast_ref, func_nd_bcast_ip)])

func_nd_params = [(func_nd_ref, f) for f in [func_nd_oop]]
func_nd_params.extend([(func_nd_bcast_ref, func_nd_bcast_oop)])

func_nd = simple_fixture('func_nd', func_nd_params,
                         fmt=' {name} = {value[1].__name__} ')


def func_nd_other(x):
    return sum(x) + 1


def func_param_nd_oop(x, c):
    return sum(x) + c


def func_param_nd_ip(x, out, c):
    out[:] = sum(x) + c


def func_param_switched_nd_ip(x, c, out):
    out[:] = sum(x) + c


def func_param_bcast_nd_ref(x, c):
    return x[0] + c + 0 * sum(x[1:])


def func_param_bcast_nd_oop(x, c):
    return x[0] + c


def func_param_bcast_nd_ip(x, out, c):
    out[:] = x[0] + c


func_param_nd_ref = func_param_nd_oop
# func_param_nd_params = [(func_param_nd_ref, f)
#                         for f in [func_param_nd_oop, func_param_nd_ip,
#                                   func_param_switched_nd_ip]]
# func_param_nd_params.extend(
#     [(func_param_bcast_nd_ref, func_param_bcast_nd_oop),
#      (func_param_bcast_nd_ref, func_param_bcast_nd_ip)])

func_param_nd_params = [(func_param_nd_ref, f)
                        for f in [func_param_nd_oop]]
func_param_nd_params.extend(
    [(func_param_bcast_nd_ref, func_param_bcast_nd_oop)])

func_param_nd = simple_fixture('func_with_param', func_param_nd_params,
                               fmt=' {name} = {value[1].__name__} ')


def func_1d_ref(x):
    return x[0] * 2


def func_1d_oop(x):
    return x * 2


def func_1d_ip(x, out):
    out[:] = x * 2


func_1d_params = [
    (func_1d_ref, func_1d_oop),
    (func_1d_ref, func_1d_ip),
    (lambda x: -x[0], np.negative),
]
func_1d = simple_fixture('func_1d', func_1d_params,
                         fmt=' {name} = {value[1].__name__} ')


def func_complex_nd_oop(x):
    return sum(x) + 1j


# --- pytest fixtures (vector-valued test functions) --- #


def func_vec_nd_ref(x):
    return [sum(x) + 1, sum(x) - 1]


def func_vec_nd_oop(x):
    return (sum(x) + 1, sum(x) - 1)


func_nd_oop_seq = FuncList([lambda x: sum(x) + 1, lambda x: sum(x) - 1])
func_nd_oop_seq.__name__ = 'func_nd_oop_seq'


def func_vec_nd_ip(x, out):
    out[0] = sum(x) + 1
    out[1] = sum(x) - 1


def comp0_nd(x, out):
    out[:] = sum(x) + 1


def comp1_nd(x, out):
    out[:] = sum(x) - 1


def func_vec_nd_dual(x, out=None):
    if out is None:
        return (sum(x) + 1, sum(x) - 1)
    else:
        out[0] = sum(x) + 1
        out[1] = sum(x) - 1


func_nd_ip_seq = FuncList([comp0_nd, comp1_nd])
func_nd_ip_seq.__name__ = 'func_nd_ip_seq'

func_vec_nd_params = [(func_vec_nd_ref, f)
                      for f in [func_vec_nd_oop, func_nd_oop_seq,]]
func_vec_nd = simple_fixture('func_vec_nd', func_vec_nd_params,
                             fmt=' {name} = {value[1].__name__} ')


def func_vec_nd_other(x):
    x, backend = get_array_and_backend(x)
    return backend.array_constructor([sum(x) + 2, sum(x) + 3], device=x.device)



def func_vec_1d_ref(x):
    x, backend = get_array_and_backend(x)
    return backend.array_constructor([x[0] * 2, x[0] + 1], device=x.device)


def func_vec_1d_oop(x):
    return (x * 2, x + 1)


func_1d_oop_seq = FuncList([lambda x: x * 2, lambda x: x + 1])
func_1d_oop_seq.__name__ = 'func_1d_oop_seq'


def func_vec_1d_ip(x, out):
    out[0] = x * 2
    out[1] = x + 1


def comp0_1d(x, out):
    out[:] = x * 2


def comp1_1d(x, out):
    out[:] = x + 1


func_1d_ip_seq = FuncList([comp0_1d, comp1_1d])
func_1d_ip_seq.__name__ = 'func_1d_ip_seq'

func_vec_1d_params = [(func_vec_1d_ref, f)
                      for f in [func_vec_1d_oop, func_1d_oop_seq,
                                func_vec_1d_ip, func_1d_ip_seq]]
func_vec_1d = simple_fixture('func_vec_1d', func_vec_1d_params,
                             fmt=' {name} = {value[1].__name__} ')


def func_vec_complex_nd_oop(x):
    return (sum(x) + 1j, sum(x) - 1j)


# --- pytest fixtures (tensor-valued test functions) --- #


def func_tens_ref(x):
    # Reference function where all shapes in the list are correct
    # without broadcasting
    shp = np.broadcast(*x).shape
    return np.array([[x[0] - x[1], np.zeros(shp), x[1] + 0 * x[0]],
                     [np.ones(shp), x[0] + 0 * x[1], sum(x)]])


def func_tens_oop(x):
    # Output shape 2x3, input 2-dimensional. Broadcasting supported.
    return [[x[0] - x[1], 0, x[1]],
            [1, x[0], sum(x)]]


def func_tens_ip(x, out):
    # In-place version
    out[0, 0] = x[0] - x[1]
    out[0, 1] = 0
    out[0, 2] = x[1]
    out[1, 0] = 1
    out[1, 1] = x[0]
    out[1, 2] = sum(x)


# Array of functions. May contain constants. Should yield the same as func.
func_tens_oop_seq = FuncList([[lambda x: x[0] - x[1], 0, lambda x: x[1]],
                              [1, lambda x: x[0], lambda x: sum(x)]])
func_tens_oop_seq.__name__ = 'func_tens_oop_seq'


# In-place component functions, cannot use lambdas
def comp00(x, out):
    out[:] = x[0] - x[1]


def comp01(x, out):
    out[:] = 0


def comp02(x, out):
    out[:] = x[1]


def comp10(x, out):
    out[:] = 1


def comp11(x, out):
    out[:] = x[0]


def comp12(x, out):
    out[:] = sum(x)


func_tens_ip_seq = FuncList([[comp00, comp01, comp02],
                             [comp10, comp11, comp12]])
func_tens_ip_seq.__name__ = 'func_tens_ip_seq'


def func_tens_dual(x, out=None):
    if out is None:
        return [[x[0] - x[1], 0, x[1]],
                [1, x[0], sum(x)]]
    else:
        out[0, 0] = x[0] - x[1]
        out[0, 1] = 0
        out[0, 2] = x[1]
        out[1, 0] = 1
        out[1, 1] = x[0]
        out[1, 2] = sum(x)


func_tens_params = [(func_tens_ref, f)
                    for f in [func_tens_oop, func_tens_oop_seq,
                              func_tens_ip, func_tens_ip_seq]]
func_tens = simple_fixture('func_tens', func_tens_params,
                           fmt=' {name} = {value[1].__name__} ')


def func_tens_other(x):
    return np.array([[x[0] + x[1], sum(x), sum(x)],
                     [sum(x), 2 * x[0] - x[1], sum(x)]])


def func_tens_complex_oop(x):
    return [[x[0], 0, 1j * x[0]],
            [1j, x, sum(x) + 1j]]


# --- point_collocation tests --- #


# def test_point_collocation_scalar_valued(domain_ndim, out_dtype, func_nd):
#     """Check collocation of scalar-valued functions."""
#     domain = odl.IntervalProd([0] * domain_ndim, [1] * domain_ndim)
#     points = _points(domain, 3)
#     mesh_shape = tuple(range(2, 2 + domain_ndim))
#     mesh = _meshgrid(domain, mesh_shape)
#     point = [0.5] * domain_ndim

#     func_ref, func = func_nd

#     true_values_points = func_ref(points)
#     true_values_mesh = func_ref(mesh)
#     true_value_point = func_ref(point)

#     sampl_func = sampling_function(func, domain, out_dtype)
#     collocator = partial(point_collocation, sampl_func)

#     # Out of place
#     result_points = collocator(points) 
#     result_mesh = collocator(mesh)
#     assert all_almost_equal(result_points, true_values_points)
#     assert all_almost_equal(result_mesh, true_values_mesh)
#     # assert result_points.flags.writeable
#     # assert result_mesh.flags.writeable

#     # In place: NOT SUPPORTED ANYMORE
#     # out_points = np.empty(3)
#     # out_mesh = np.empty(mesh_shape)
#     # collocator(points, out=out_points)
#     # collocator(mesh, out=out_mesh)

#     # assert all_almost_equal(out_points, true_values_points)
#     # assert all_almost_equal(out_mesh, true_values_mesh)

#     # Single point evaluation
#     result_point = collocator(point)
#     assert all_almost_equal(result_point, true_value_point)


def test_point_collocation_scalar_valued_with_param(odl_impl_device_pairs):
    """Check collocation of scalar-valued functions with parameters."""
    domain = odl.IntervalProd([0, 0], [1, 1])
    points = _points(domain, 3)
    mesh_shape = (2, 3)
    mesh = _meshgrid(domain, mesh_shape)

    impl, device = odl_impl_device_pairs
    backend = lookup_array_backend(impl)

    def func_ref(x, c):
        if isinstance(x, (tuple, list)):
            return [func_ref(x_, c) for x_ in x]
        
        return np.sum(x) + c

    def func(x, c):
        if isinstance(x, (tuple, list)):
            return [func(x_, c) for x_ in x]
        
        return backend.array_namespace.sum(x) + c

    true_values_points = backend.array_constructor(
        func_ref(points, c=2.5), device=device)
    true_values_mesh = backend.array_constructor(
        func_ref(mesh, c=2.5), device=device) 

    sampl_func = sampling_function(func, domain, out_dtype='float64', impl=impl, device=device)
    collocator = partial(point_collocation, sampl_func)

    # Out of place
    result_points = collocator(points, c=2.5)
    result_mesh = collocator(mesh, c=2.5)
    assert all_almost_equal(result_points, true_values_points)
    assert all_almost_equal(result_mesh, true_values_mesh)

    # Complex output
    sampl_func = sampling_function(func, domain, out_dtype='complex128', impl=impl, device=device)
    collocator = partial(point_collocation, sampl_func)

    true_values_points = backend.array_constructor(
        func_ref(points, c=2j), device=device)
    true_values_mesh = backend.array_constructor(
        func_ref(mesh, c=2j), device=device)

    result_points = collocator(points, c=2j)
    result_mesh = collocator(mesh, c=2j)
    assert all_almost_equal(result_points, true_values_points)
    assert all_almost_equal(result_mesh, true_values_mesh)


def test_point_collocation_vector_valued(odl_impl_device_pairs):
    """Check collocation of vector-valued functions."""
    domain = odl.IntervalProd([0, 0], [1, 1])
    points = _points(domain, 3)
    mesh_shape = (2, 3)
    mesh = _meshgrid(domain, mesh_shape)
    point = [0.5, 0.5]

    impl, device = odl_impl_device_pairs
    backend = lookup_array_backend(impl)
    ns = backend.array_namespace
    def func_vec_nd_ref(x):
        return (np.sin(x[0])+ np.sin(x[1])+ 1, np.sin(x[0])+np.sin(x[1]) -1 )
    def func_vec_nd_torch(x):
        return (
            ns.sin(x[0])+ ns.sin(x[1])+ 1,  
            ns.sin(x[0])+ ns.sin(x[1])- 1 
        )
    
    impl, device = odl_impl_device_pairs
    func_ref = func_vec_nd_ref
    if impl == 'pytorch':
        func = func_vec_nd_torch
    else:
        func = func_vec_nd_ref       

    true_values_points = func_ref(points)
    true_values_mesh = func_ref(mesh)
    true_value_point = func_ref(point)

    backend = lookup_array_backend(impl)
    true_values_points = backend.array_constructor(func_ref(points), device=device)
    true_values_mesh = backend.array_constructor(func_ref(mesh), device=device)
    true_value_point = backend.array_constructor(func_ref(point), device=device)

    sampl_func = sampling_function(
        func, domain, out_dtype='float64', impl=impl, device=device
    )
    collocator = partial(point_collocation, sampl_func)

    # Out of place
    result_points = collocator(points)
    result_mesh = collocator(mesh)

    assert all_almost_equal(result_points, true_values_points)
    assert all_almost_equal(result_mesh, true_values_mesh)

    # Single point evaluation
    result_point = collocator(point)
    assert all_almost_equal(result_point, true_value_point)

# def test_point_collocation_tensor_valued():
#     """Check collocation of tensor-valued functions."""
#     domain = odl.IntervalProd([0, 0], [1, 1])
#     points = _points(domain, 4)
#     mesh_shape = (4, 5)
#     mesh = _meshgrid(domain, mesh_shape)
#     point = [0.5, 0.5]
#     values_points_shape = (2, 3, 4)
#     values_mesh_shape = (2, 3, 4, 5)
#     value_point_shape = (2, 3)

#     def func_tens_oop(x):
#         # Output shape 2x3, input 2-dimensional. Broadcasting supported.
#         return [[x[0] - x[1], 0, x[1]],
#                 [1, x[0], sum(x)]]

#     func_ref = func_tens_oop
#     func = func_tens_oop

#     true_result_points = np.array(func_ref(points))
#     true_result_mesh = np.array(func_ref(mesh))
#     true_result_point = np.array(func_ref(np.array(point)[:, None])).squeeze()

#     sampl_func = sampling_function(
#         func, domain, out_dtype='float64'
#     )
#     collocator = partial(point_collocation, sampl_func)

#     result_points = collocator(points)
#     result_mesh = collocator(mesh)
#     result_point = collocator(point)
#     assert all_almost_equal(result_points, true_result_points)
#     print(result_mesh)
#     print(true_result_mesh)
#     assert all_almost_equal(result_mesh, true_result_mesh)
#     assert all_almost_equal(result_point, true_result_point)
    # assert result_points.flags.writeable
    # assert result_mesh.flags.writeable
    # assert result_point.flags.writeable

    # out_points = np.empty(values_points_shape, dtype='float64')
    # out_mesh = np.empty(values_mesh_shape, dtype='float64')
    # out_point = np.empty(value_point_shape, dtype='float64')
    # collocator(points, out=out_points)
    # collocator(mesh, out=out_mesh)
    # collocator(point, out=out_point)
    # assert all_almost_equal(out_points, true_result_points)
    # assert all_almost_equal(out_mesh, true_result_mesh)
    # assert all_almost_equal(out_point, true_result_point)


# def test_fspace_elem_eval_unusual_dtypes():
#     """Check evaluation with unusual data types (int and string)."""
#     domain = odl.Strings(3)
#     strings = np.array(['aa', 'b', 'cab', 'aba'])
#     out_vec = np.empty((4,), dtype='int64')

#     # Can be vectorized for arrays only
#     sampl_func = sampling_function(
#         lambda s: np.array([str(si).count('a') for si in s]),
#         domain,
#         out_dtype='int64'
#     )
#     collocator = partial(point_collocation, sampl_func)

#     true_values = [2, 0, 1, 2]

#     assert collocator('abc') == 1
#     assert all_equal(collocator(strings), true_values)
#     collocator(strings, out=out_vec)
#     assert all_equal(out_vec, true_values)


# def test_fspace_elem_eval_vec_1d(func_vec_1d):
#     """Test evaluation in 1d since it's a corner case regarding shapes."""
#     domain = odl.IntervalProd(0, 1)
#     points = _points(domain, 3)
#     mesh_shape = (4,)
#     mesh = _meshgrid(domain, mesh_shape)
#     point1 = 0.5
#     point2 = [0.5]
#     values_points_shape = (2, 3)
#     values_mesh_shape = (2, 4)
#     value_point_shape = (2,)

#     func_ref, func = func_vec_1d

#     true_result_points = np.array(func_ref(points))
#     true_result_mesh = np.array(func_ref(mesh))
#     true_result_point = np.array(func_ref(np.array([point1]))).squeeze()

#     sampl_func = sampling_function(
#         func, domain, out_dtype=('float64', (2,))
#     )
#     collocator = partial(point_collocation, sampl_func)

#     result_points = collocator(points)
#     result_mesh = collocator(mesh)
#     result_point1 = collocator(point1)
#     result_point2 = collocator(point2)
#     assert all_almost_equal(result_points, true_result_points)
#     assert all_almost_equal(result_mesh, true_result_mesh)
#     assert all_almost_equal(result_point1, true_result_point)
#     assert all_almost_equal(result_point2, true_result_point)

#     out_points = np.empty(values_points_shape, dtype='float64')
#     out_mesh = np.empty(values_mesh_shape, dtype='float64')
#     out_point1 = np.empty(value_point_shape, dtype='float64')
#     out_point2 = np.empty(value_point_shape, dtype='float64')
#     collocator(points, out=out_points)
#     collocator(mesh, out=out_mesh)
#     collocator(point1, out=out_point1)
#     collocator(point2, out=out_point2)
#     assert all_almost_equal(out_points, true_result_points)
#     assert all_almost_equal(out_mesh, true_result_mesh)
#     assert all_almost_equal(out_point1, true_result_point)
#     assert all_almost_equal(out_point2, true_result_point)


# --- interpolation tests --- #

def test_nearest_interpolation_1d_complex(odl_impl_device_pairs):
    """Test nearest neighbor interpolation in 1d with complex values."""
    coord_vecs = [[0.1, 0.3, 0.5, 0.7, 0.9]]

    impl, device = odl_impl_device_pairs
    backend = lookup_array_backend(impl)
    if impl == 'pytorch':
        pytest.skip('Interpolator class not implemented for pytorch complex dtypes')
    dtype = backend.available_dtypes["complex128"]
    f = backend.array_constructor(
        [0 + 1j, 1 + 2j, 2 + 3j, 3 + 4j, 4 + 5j], 
        dtype=dtype, device=device
        )
    
    interpolator = nearest_interpolator(f, coord_vecs)

    # Evaluate at single point
    val = interpolator(0.35)  # closest to index 1 -> 1 + 2j
    assert val == 1.0 + 2.0j
    # Input array, with and without output array
    pts = np.array([0.39, 0.0, 0.65, 0.95])
    true_arr = [1 + 2j, 0 + 1j, 3 + 4j, 4 + 5j]
    assert all_equal(interpolator(pts), true_arr)
    # Should also work with a (1, N) array
    pts = pts[None, :]
    assert all_equal(interpolator(pts), true_arr)
    out = backend.array_namespace.empty(4, dtype=dtype, device=device)
    interpolator(pts, out=out)
    assert all_equal(out, true_arr)
    # Input meshgrid, with and without output array
    # Same as array for 1d
    mg = sparse_meshgrid([0.39, 0.0, 0.65, 0.95])
    true_mg = [1 + 2j, 0 + 1j, 3 + 4j, 4 + 5j]
    assert all_equal(interpolator(mg), true_mg)
    interpolator(mg, out=out)
    assert all_equal(out, true_mg)


def test_nearest_interpolation_2d(odl_impl_device_pairs):
    """Test nearest neighbor interpolation in 2d."""
    coord_vecs = [[0.125, 0.375, 0.625, 0.875], [0.25, 0.75]]

    impl, device = odl_impl_device_pairs
    backend = lookup_array_backend(impl)
    dtype = backend.available_dtypes["float64"]
    
    f = backend.array_constructor([[0, 1],
                  [2, 3],
                  [4, 5],
                  [6, 7]], dtype=dtype, device=device)
    interpolator = nearest_interpolator(f, coord_vecs)

    # Evaluate at single point
    val = interpolator([0.3, 0.6])  # closest to index (1, 1) -> 3
    assert val == 3.0
    # Input array, with and without output array
    pts = np.array([[0.3, 0.6],
                    [1.0, 1.0]])
    true_arr = [3, 7]
    assert all_equal(interpolator(pts.T), true_arr)
    out = backend.array_namespace.empty(2, dtype=dtype, device=device)
    interpolator(pts.T, out=out)
    assert all_equal(out, true_arr)
    # Input meshgrid, with and without output array
    mg = sparse_meshgrid([0.3, 1.0], [0.4, 1.0])
    # Indices: (1, 3) x (0, 1)
    true_mg = [[2, 3],
               [6, 7]]
    assert all_equal(interpolator(mg), true_mg)
    out = backend.array_namespace.empty((2, 2), dtype=dtype, device=device)
    interpolator(mg, out=out)
    assert all_equal(out, true_mg)


def test_linear_interpolation_1d(odl_impl_device_pairs):
    """Test linear interpolation in 1d."""
    coord_vecs = [[0.1, 0.3, 0.5, 0.7, 0.9]]
    impl, device = odl_impl_device_pairs
    backend = lookup_array_backend(impl)
    dtype = backend.available_dtypes["float64"]
    
    f = backend.array_constructor(
        [1, 2, 3, 4, 5], dtype=dtype, device=device
        )
    
    interpolator = linear_interpolator(f, coord_vecs)

    # Evaluate at single point
    val = interpolator(0.35)
    true_val = 0.75 * 2 + 0.25 * 3
    assert val == pytest.approx(true_val)

    # Input array, with and without output array
    pts = np.array([0.4, 0.0, 0.65, 0.95])
    true_arr =  backend.array_constructor([2.5, 0.5, 3.75, 3.75], dtype=dtype, device=device)
    assert all_almost_equal(interpolator(pts), true_arr)


def test_linear_interpolation_2d(odl_impl_device_pairs):
    """Test linear interpolation in 2d."""
    coord_vecs = [[0.125, 0.375, 0.625, 0.875], [0.25, 0.75]]

    impl, device = odl_impl_device_pairs
    backend = lookup_array_backend(impl)
    dtype = backend.available_dtypes["float64"]

    f = backend.array_constructor(
            [[1, 2],
            [3, 4],
            [5, 6],
            [7, 8]], dtype=dtype, device=device
        )
    
    interpolator = linear_interpolator(f, coord_vecs)

    # Evaluate at single point
    val = interpolator([0.3, 0.6])
    l1 = (0.3 - 0.125) / (0.375 - 0.125)
    l2 = (0.6 - 0.25) / (0.75 - 0.25)
    true_val = (
        (1 - l1) * (1 - l2) * f[0, 0]
        + (1 - l1) * l2 * f[0, 1]
        + l1 * (1 - l2) * f[1, 0]
        + l1 * l2 * f[1, 1]
    )
    if impl == 'pytorch':
        true_val = true_val.detach().cpu()
    assert val == pytest.approx(true_val)

    # Input array, with and without output array
    pts = np.array([[0.3, 0.6],
                    [0.1, 0.25],
                    [1.0, 1.0]])
    l1 = (0.3 - 0.125) / (0.375 - 0.125)
    l2 = (0.6 - 0.25) / (0.75 - 0.25)
    true_val_1 = (
        (1 - l1) * (1 - l2) * f[0, 0]
        + (1 - l1) * l2 * f[0, 1]
        + l1 * (1 - l2) * f[1, 0]
        + l1 * l2 * f[1, 1]
    )
    l1 = (0.125 - 0.1) / (0.375 - 0.125)
    # l2 = 0
    true_val_2 = (1 - l1) * f[0, 0]  # only lower left contributes
    l1 = (1.0 - 0.875) / (0.875 - 0.625)
    l2 = (1.0 - 0.75) / (0.75 - 0.25)
    true_val_3 = (1 - l1) * (1 - l2) * f[3, 1]  # lower left only
    true_arr = [true_val_1, true_val_2, true_val_3]
    assert all_equal(interpolator(pts.T), true_arr)

    out = backend.array_namespace.empty(3, dtype=dtype, device=device)
    interpolator(pts.T, out=out)
    assert all_equal(out, true_arr)

    # Input meshgrid, with and without output array
    mg = sparse_meshgrid([0.3, 1.0], [0.4, 0.75])
    # Indices: (1, 3) x (0, 1)
    lx1 = (0.3 - 0.125) / (0.375 - 0.125)
    lx2 = (1.0 - 0.875) / (0.875 - 0.625)
    ly1 = (0.4 - 0.25) / (0.75 - 0.25)
    # ly2 = 0
    true_val_11 = (
        (1 - lx1) * (1 - ly1) * f[0, 0]
        + (1 - lx1) * ly1 * f[0, 1]
        + lx1 * (1 - ly1) * f[1, 0]
        + lx1 * ly1 * f[1, 1]
    )
    true_val_12 = (
        (1 - lx1) * f[0, 1]
        + lx1 * f[1, 1]  # ly2 = 0
    )
    true_val_21 = (
        (1 - lx2) * (1 - ly1) * f[3, 0]
        + (1 - lx2) * ly1 * f[3, 1]   # high node 1.0, no upper
    )
    true_val_22 = (1 - lx2) * f[3, 1]  # ly2 = 0, no upper for 1.0
    true_mg = [[true_val_11, true_val_12],
               [true_val_21, true_val_22]]
    assert all_equal(interpolator(mg), true_mg)
    out = backend.array_namespace.empty((2, 2), dtype=dtype, device=device)
    interpolator(mg, out=out)
    assert all_equal(out, true_mg)


def test_per_axis_interpolation(odl_impl_device_pairs):
    """Test different interpolation schemes per axis."""
    coord_vecs = [[0.125, 0.375, 0.625, 0.875], [0.25, 0.75]]
    interp = ['linear', 'nearest']
    impl, device = odl_impl_device_pairs
    backend = lookup_array_backend(impl)
    dtype = backend.available_dtypes["float64"]

    f = backend.array_constructor(
            [[1, 2],
            [3, 4],
            [5, 6],
            [7, 8]], dtype=dtype, device=device
        )
    interpolator = per_axis_interpolator(f, coord_vecs, interp)

    # Evaluate at single point
    val = interpolator([0.3, 0.5])
    l1 = (0.3 - 0.125) / (0.375 - 0.125)
    # 0.5 equally far from both neighbors -> NN chooses 0.75
    true_val = (1 - l1) * f[0, 1] + l1 * f[1, 1]
    if impl == 'pytorch':
        true_val = true_val.detach().cpu()
    assert val == pytest.approx(true_val)

    # Input array, with and without output array
    pts = np.array([[0.3, 0.6],
                    [0.1, 0.25],
                    [1.0, 1.0]])
    l1 = (0.3 - 0.125) / (0.375 - 0.125)
    true_val_1 = (1 - l1) * f[0, 1] + l1 * f[1, 1]
    l1 = (0.125 - 0.1) / (0.375 - 0.125)
    true_val_2 = (1 - l1) * f[0, 0]  # only lower left contributes
    l1 = (1.0 - 0.875) / (0.875 - 0.625)
    true_val_3 = (1 - l1) * f[3, 1]  # lower left only
    true_arr = [true_val_1, true_val_2, true_val_3]
    assert all_equal(interpolator(pts.T), true_arr)

    out = backend.array_namespace.empty(3, dtype=dtype, device=device)
    interpolator(pts.T, out=out)
    assert all_equal(out, true_arr)

    # Input meshgrid, with and without output array
    mg = sparse_meshgrid([0.3, 1.0], [0.4, 0.85])
    # Indices: (1, 3) x (0, 1)
    lx1 = (0.3 - 0.125) / (0.375 - 0.125)
    lx2 = (1.0 - 0.875) / (0.875 - 0.625)
    true_val_11 = (1 - lx1) * f[0, 0] + lx1 * f[1, 0]
    true_val_12 = ((1 - lx1) * f[0, 1] + lx1 * f[1, 1])
    true_val_21 = (1 - lx2) * f[3, 0]
    true_val_22 = (1 - lx2) * f[3, 1]
    true_mg = [[true_val_11, true_val_12],
               [true_val_21, true_val_22]]
    assert all_equal(interpolator(mg), true_mg)
    out = backend.array_namespace.empty((2, 2), dtype=dtype, device=device)
    interpolator(mg, out=out)
    assert all_equal(out, true_mg)


def test_collocation_interpolation_identity(odl_impl_device_pairs):
    """Check if collocation is left-inverse to interpolation."""
    # Interpolation followed by collocation on the same grid should be
    # the identity
    coord_vecs = [[0.125, 0.375, 0.625, 0.875], [0.25, 0.75]]
    impl, device = odl_impl_device_pairs
    backend = lookup_array_backend(impl)
    dtype = backend.available_dtypes["float64"]

    f = backend.array_constructor(
            [[1, 2],
            [3, 4],
            [5, 6],
            [7, 8]], dtype=dtype, device=device
        )
    interpolators = [
        nearest_interpolator(f, coord_vecs),
        linear_interpolator(f, coord_vecs),
        per_axis_interpolator(f, coord_vecs, interp=['linear', 'nearest']),
    ]

    for interpolator in interpolators:
        mg = sparse_meshgrid(*coord_vecs)
        ident_f = point_collocation(interpolator, mg)
        assert all_almost_equal(ident_f, f)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2018 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Step length computation for optimization schemes."""

from __future__ import print_function, division, absolute_import
from builtins import object
import numpy as np


__all__ = ('LineSearch', 'BacktrackingLineSearch', 'ConstantLineSearch',
           'LineSearchFromIterNum')


class LineSearch(object):

    """Abstract base class for line search step length methods."""

    def __call__(self, x, direction, dir_derivative):
        """Calculate step length in direction.

        Parameters
        ----------
        x : `LinearSpaceElement`
            The current point
        direction : `LinearSpaceElement`
            Search direction in which the line search should be computed
        dir_derivative : float
            Directional derivative along the ``direction``

        Returns
        -------
        step : float
            Computed step length.
        """
        raise NotImplementedError('abstract method')


class BacktrackingLineSearch(LineSearch):

    """Backtracking line search for step length calculation.

    This methods approximately finds the longest step length fulfilling
    the Armijo-Goldstein condition.

    The line search algorithm is described in [BV2004], page 464
    (`book available online
    <http://stanford.edu/~boyd/cvxbook/bv_cvxbook.pdf>`_) and
    [GNS2009], pages 378--379. See also
    `Backtracking_line_search
    <https://en.wikipedia.org/wiki/Backtracking_line_search>`_.

    References
    ----------
    [BV2004] Boyd, S, and Vandenberghe, L. *Convex optimization*.
    Cambridge university press, 2004.

    [GNS2009] Griva, I, Nash, S G, and Sofer, A. *Linear and nonlinear
    optimization*. Siam, 2009.
    """

    def __init__(self, function, tau=0.5, discount=0.01, alpha=1.0,
                 max_num_iter=None, estimate_step=False):
        """Initialize a new instance.

        Parameters
        ----------
        function : callable
            The cost function of the optimization problem to be solved.
            If ``function`` is not a `Functional`, calling this class later
            requires a value for the ``dir_derivative`` argument.
        tau : float, optional
            The amount the step length is decreased in each iteration,
            as long as it does not fulfill the decrease condition.
            The step length is updated as ``step_length *= tau``.
        discount : float, optional
            The "discount factor" on ``step length * direction derivative``,
            yielding the threshold under which the function value must lie to
            be accepted (see the references).
        alpha : float, optional
            The initial guess for the step length.
        max_num_iter : int, optional
            Maximum number of iterations allowed each time the line
            search method is called. If ``None``, this number is calculated
            to allow a shortest step length of 10 times machine epsilon.
        estimate_step : bool, optional
            If the last step should be used as a estimate for the next step.

        Examples
        --------
        Create line search

        >>> r3 = odl.rn(3)
        >>> func = odl.functionals.L2NormSquared(r3)
        >>> line_search = BacktrackingLineSearch(func)

        Find step in point x and direction d that decreases the function value.

        >>> x = r3.element([1, 2, 3])
        >>> d = r3.element([-1, -1, -1])
        >>> step_len = line_search(x, d)
        >>> step_len
        1.0
        >>> func(x + step_len * d) < func(x)
        True

        Also works with non-functionals as arguments, but then the
        dir_derivative argument is mandatory

        >>> r3 = odl.rn(3)
        >>> func = lambda x: x[0] ** 2 + x[1] ** 2 + x[2] ** 2
        >>> line_search = BacktrackingLineSearch(func)
        >>> x = r3.element([1, 2, 3])
        >>> d = r3.element([-1, -1, -1])
        >>> dir_derivative = -12
        >>> step_len = line_search(x, d, dir_derivative=dir_derivative)
        >>> step_len
        1.0
        >>> func(x + step_len * d) < func(x)
        True
        """
        self.function = function
        self.tau = float(tau)
        self.discount = float(discount)
        self.estimate_step = bool(estimate_step)
        self.alpha = float(alpha)

        self.total_num_iter = 0
        # Use a default value that allows the shortest step to be < 10 times
        # machine epsilon.
        if max_num_iter is None:
            try:
                dtype = self.function.domain.dtype_identifier
            except AttributeError:
                dtype = float
            eps = 10 * np.finfo(dtype).resolution
            self.max_num_iter = int(np.ceil(np.log(eps) / np.log(self.tau)))
        else:
            self.max_num_iter = int(max_num_iter)

    def __call__(self, x, direction, dir_derivative=None):
        """Calculate the optimal step length along a line.

        Parameters
        ----------
        x : `LinearSpaceElement`
            The current point
        direction : `LinearSpaceElement`
            Search direction in which the line search should be computed
        dir_derivative : float, optional
            Directional derivative along the ``direction``
            Default: ``function.gradient(x).inner(direction)``

        Returns
        -------
        step : float
            The computed step length
        """
        fx = self.function(x)
        if dir_derivative is None:
            try:
                gradient = self.function.gradient
            except AttributeError:
                raise ValueError('`dir_derivative` only optional if '
                                 '`function.gradient exists')
            else:
                dir_derivative = gradient(x).inner(direction)
        else:
            dir_derivative = float(dir_derivative)

        if dir_derivative == 0:
            raise ValueError('dir_derivative == 0, no descent can be found')

        if not self.estimate_step:
            alpha = 1.0
        else:
            alpha = self.alpha

        if dir_derivative > 0:
            # We need to move backwards if the direction is an increase
            # direction
            alpha *= -1

        if not np.isfinite(fx):
            raise ValueError('function returned invalid value {} in starting '
                             'point ({})'.format(fx, x))

        # Create temporary
        point = x.copy()

        num_iter = 0
        while True:
            if num_iter > self.max_num_iter:
                raise ValueError('number of iterations exceeded maximum: {}, '
                                 'step length: {}, without finding a '
                                 'sufficient decrease'
                                 ''.format(self.max_num_iter, alpha))

            point.lincomb(1, x, alpha, direction)  # pt = x + alpha * direction
            fval = self.function(point)

            if np.isnan(fval):
                # We do not want to compare against NaN below, and NaN should
                # indicate a user error.
                raise ValueError('function returned NaN in point '
                                 'point ({})'.format(point))

            expected_decrease = np.abs(alpha * dir_derivative * self.discount)
            if (fval <= fx - expected_decrease):
                # Stop iterating if the value decreases sufficiently.
                break

            num_iter += 1
            alpha *= self.tau

        assert fval < fx

        self.total_num_iter += num_iter
        self.alpha = np.abs(alpha)  # Store magnitude
        return alpha


class ConstantLineSearch(LineSearch):

    """Line search object that returns a constant step length."""

    def __init__(self, constant):
        """Initialize a new instance.

        Parameters
        ----------
        constant : float
            The constant step length
        """
        self.constant = float(constant)

    def __call__(self, x, direction, dir_derivative):
        """Calculate the step length at a point.

        All arguments are ignored and are only added to fit the interface.
        """
        return self.constant


class LineSearchFromIterNum(LineSearch):

    """Line search object that returns a step length from a function.

    The returned step length is ``func(iter_count)``.
    """

    def __init__(self, func):
        """Initialize a new instance.

        Parameters
        ----------
        func : callable
            Function that when called with an iteration count should return the
            step length. The iteration count starts at 0.

        Examples
        --------
        Make a step size that is 1.0 for the first 5 iterations, then 0.1:

        >>> def step_length(iter):
        ...     if iter < 5:
        ...         return 1.0
        ...     else:
        ...         return 0.1
        >>> line_search = LineSearchFromIterNum(step_length)
        """
        if not callable(func):
            raise TypeError('`func` must be a callable.')
        self.func = func
        self.iter_count = 0

    def __call__(self, x, direction, dir_derivative):
        """Calculate the step length at a point.

        All arguments are ignored and are only added to fit the interface.
        """
        step = self.func(self.iter_count)
        self.iter_count += 1
        return step


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()
